import os
import io
from PIL import Image
import sys
import re
import time
import numpy as np
import pygad as pg            # use absolute package loading
from .. import environment

__all__ = ["SnapshotCache", "SnapshotProperty"]

class SnapshotCache:
    def __init__(self, snapfile_name, profile='cache'):
        # type: (str, str) -> SnapshotCache
        pg.environment.verbose = environment.verbose
        if snapfile_name[0] != '/':
            snap_home = os.getenv('SNAPSHOT_HOME', '')
            if snap_home == '':
                # treat local path same as absolute path
                data_file = os.path.abspath(snapfile_name)
                cache_home = ''
            else:
                if snap_home[-1] != '/': snap_home = snap_home + '/'
                data_file = snap_home + snapfile_name
                cache_home = os.getenv('SNAPCACHE_HOME', '')
                if cache_home == '':
                    cache_home = snap_home
        else:
            data_file = snapfile_name
            snap_home = ''
            cache_home = ''

        self.__snap_home = snap_home
        self.__cache_home = cache_home
        self.__snap_name = snapfile_name
        self.__data_file = data_file
        self.__profile = ''
        self.__profile_properties = SnapshotProperty()
        self._load_profile(profile)

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("************************************************************************")
            print("using snapfile ",self.__data_file)
            print("************************************************************************")

        self.__results = dict()
        self.__snapshot = None
        self.__dm_halo = None
        self.__halo_properties = SnapshotProperty()
        self.__gx_properties = SnapshotProperty()
        self.__galaxy = None
        self.__galaxy_all = None
        self.__family = None
        self.__center = None
        self.__loaded = False
        # #########################################
        # pygad block-names
        # #########################################
        self.__B_POS  = None        # 'pos'       # POS = (3, 'float', 'all')
        self.__B_VEL  = None        # 'vel'       # VEL = (3, 'float', 'all')
        self.__B_ID   = None        # 'ID'        # ID = (1, 'uint', 'all')
        self.__B_MASS = None        # 'mass'      # MASS = (1, 'float', None)
        self.__B_U    = None        # 'u'         # U = (1, 'float', 'gas')
        self.__B_RHO  = None        # 'rho'       # RHO = (1, 'float', 'gas')
        self.__B_HSML = None        # 'hsml'      # HSML = (1, 'float', 'gas')
        self.__B_AGE  = None        # 'age'       # AGE = (1, 'float', None)

        # ##########################################
        # derived block-names (always cached age, temp, mag*, Ekin, angmom, jcirc, LX
        # ##########################################
        self.__D_P    = None        # 'P'          pressure
        self.__D_temp = None        # 'temp'
        self.__D_Ekin = None        # 'Ekin'
        self.__D_Epot = None        # 'Epot'
        self.__D_E    = None        # 'E'
        self.__D_r    = None        # 'r'         # spherical radius
        self.__D_rcyl = None        # 'rcyl'      # dist(pos[:, :2]); cylindrical radius
        self.__D_vrad = None        # 'vrad'      # inner1d(pos, vel) / r;  radial velocities
        self.__D_momentum = None        # 'momentum' # (mass * vel.T).T; momentum
        self.__D_angmom   = None        # 'angmom'   # UnitQty(cross(pos, momentum), pos.units * momentum.units); angular momentum
        # the angular momentum a particle on a circular orbit with the same energy would have
        self.__D_jcirc = None        # 'jcirc'    # r * sqrt(2.0 * mass * Ekin)
        # the parameter jz/jc, where jz is the z-component of the angular momentum and jc=jcirc:
        self.__D_jzjc = None        # 'jzjc'      # angmom[:, 2] / jcirc
        self.__D_vcirc= None        # 'vcirc'      # sqrt(sum(vel ** 2, axis=-1) - vrad ** 2); circular part of the velocities
        # derived from tracing files:
        self.__D_insitu = None        # 'insitu'  # rR200form < 0.1

        self.__L_lum   = None        # 'lum'
        self.__L_lum_u = None        # 'lum_u'
        self.__L_lum_b = None        # 'lum_b'
        self.__L_lum_v = None        # 'lum_v'
        self.__L_lum_r = None        # 'lum_r'
        self.__L_lum_k = None        # 'lum_k'
        self.__L_LX    = None        # 'LX'
        self.__L_mag   = None        # 'mag'
        self.__L_mag_u = None        # 'mag_u'
        self.__L_mag_b = None        # 'mag_b'
        self.__L_mag_v = None        # 'mag_v'
        self.__L_mag_r = None        # 'mag_r'
        self.__L_mag_k = None        # 'mag_k'

    @property
    def B_POS(self):
        # type: () -> pg.SimArr
        return self.galaxy['pos']

    @property
    def B_VEL(self):
        # type: () -> pg.SimArr
        return self.__galaxy['vel']

    @property
    def B_ID(self):
        # type: () -> pg.SimArr
        return self.__galaxy['ID']

    @property
    def B_MASS(self):
        # type: () -> pg.SimArr
        return self.__galaxy['mass']

    @property
    def B_U(self):
        # type: () -> pg.SimArr
        return self.__galaxy['u']

    @property
    def B_RHO(self):
        # type: () -> pg.SimArr
        return self.__galaxy['rho']

    @property
    def B_HSML(self):
        # type: () -> pg.SimArr
        return self.__galaxy['hsml']

    @property
    def B_AGE(self):
        # type: () -> pg.SimArr
        return self.__galaxy['age']

    @property
    def D_P(self):
        # type: () -> pg.SimArr
        return self.__galaxy['P']

    @property
    def D_temp(self):
        # type: () -> pg.SimArr
        return self.__galaxy['temp']

    @property
    def D_Ekin(self):
        # type: () -> pg.SimArr
        return self.__galaxy['Ekin']

    @property
    def D_Epot(self):
        # type: () -> pg.SimArr
        return self.__galaxy['Epot']

    @property
    def D_E(self):
        # type: () -> pg.SimArr
        return self.__galaxy['E']

    @property
    def D_r(self):
        # type: () -> pg.SimArr
        return self.__galaxy['r']

    @property
    def D_rcyl(self):
        # type: () -> pg.SimArr
        return self.__galaxy['rcyl']

    @property
    def D_vrad(self):
        # type: () -> pg.SimArr
        return self.__galaxy['vrad']

    @property
    def D_momentum(self):
        # type: () -> pg.SimArr
        return self.__galaxy['momentum']

    @property
    def D_angmom(self):
        # type: () -> pg.SimArr
        return self.__galaxy['agmom']

    @property
    def D_jcirc(self):
        # type: () -> pg.SimArr
        return self.__galaxy['jcirc']

    @property
    def D_jzjc(self):
        # type: () -> pg.SimArr
        return self.__galaxy['jzjc']

    @property
    def D_vcirc(self):
        # type: () -> pg.SimArr
        return self.__galaxy['vcirc']

    @property
    def D_insitu(self):
        # type: () -> pg.SimArr
        return self.__galaxy['insitu']

    @property
    def L_lum(self):
        # type: () -> pg.SimArr
        return self.__galaxy['lum']

    @property
    def L_lum_u(self):
        # type: () -> pg.SimArr
        return self.__galaxy['lum_u']

    @property
    def L_lum_b(self):
        # type: () -> pg.SimArr
        return self.galaxy['lum_b']
    @property
    def L_lum_v(self):
        # type: () -> pg.SimArr
        return self.__galaxy['lum_v']

    @property
    def L_lum_r(self):
        # type: () -> pg.SimArr
        return self.__galaxy['lum_r']

    @property
    def L_lum_k(self):
        # type: () -> pg.SimArr
        return self.__galaxy['lum_k']

    @property
    def D_LX(self):
        # type: () -> pg.SimArr
        return self.__galaxy['LX']

    @property
    def L_mag(self):
        # type: () -> pg.SimArr
        return self.__galaxy['mag']

    @property
    def L_mag_u(self):
        # type: () -> pg.SimArr
        return self.__galaxy['mag_u']

    @property
    def L_mag_b(self):
        # type: () -> pg.SimArr
        return self.__galaxy['mag_b']

    @property
    def L_mag_v(self):
        # type: () -> pg.SimArr
        return self.__galaxy['mag_v']

    @property
    def L_mag_r(self):
        # type: () -> pg.SimArr
        return self.__galaxy['mag_r']

    @property
    def L_mag_k(self):
        # type: () -> pg.SimArr
        return self.__galaxy['mag_k']

    # ##########################################
    # Data processing properties and methods
    # ##########################################

    @property
    def name(self):
        return self.__data_file

    @property
    def halo(self):
        if self.__dm_halo is None:
            Rvir_property = self.__profile_properties['Rvir_property']
            if Rvir_property in self.__halo_properties:
                Rvir = self.__halo_properties[Rvir_property]
                mask = pg.BallMask(Rvir, center=[0, 0, 0])
                halo = self.__snapshot[mask]
                self.__dm_halo = halo
        return self.__dm_halo

    @property
    def galaxy(self):
        # type: () -> pg.SubSnapshot
        return self.__galaxy

    @property
    def family(self):
        return self.__family

    @family.setter
    def family(self, value):
        if value is None:
            self.__galaxy = self.__galaxy_all
        else:
            gx = self.__galaxy_all
            self.__galaxy = eval('gx.' + str(value),globals(),locals())

        self.__family = value

    @property
    def gas(self):
        # type: () -> pg.SubSnapshot
        return self.__galaxy_all.gas

    @property
    def stars(self):
        # type: () -> pg.SubSnapshot
        return self.__galaxy_all.stars

    @property
    def dm(self):
        # type: () -> pg.SubSnapshot
        return self.__galaxy_all.dm

    @property
    def bh(self):
        # type: () -> pg.SubSnapshot
        return self.__galaxy_all.bh

    @property
    def gands(self):
        # type: () -> pg.SubSnapshot
        return self.__galaxy_all.gands

    @property
    def sandbh(self):
        # type: () -> pg.SubSnapshot
        return self.__galaxy_all.sandbh
    @property

    def baryons(self):
        # type: () -> pg.SubSnapshot
        return self.__galaxy_all.baryons

    def highres(self):
        # type: () -> pg.SubSnapshot
        return self.__galaxy_all.highres

    def lowres(self):
        # type: () -> pg.SubSnapshot
        return self.__galaxy_all.lowres

    @property
    def halo_properties(self):
        return self.__halo_properties

    @property
    def gx_properties(self):
        return self.__gx_properties

    @property
    def profile_properties(self):
        return self.__profile_properties

    @property
    def profile(self):
        profile = self.__profile
        return profile

    @property
    def snapshot(self):
        return self.__snapshot

    @snapshot.setter
    def snapshot(self, value):
        self.__snapshot = value

    @property
    def results(self):
        return self.__results

    @property
    def Rvir_property(self):
        return self.__profile_properties['Rvir_property']

    @Rvir_property.setter
    def Rvir_property(self, value):
        self.__profile_properties.append('Rvir_property', value)

    @property
    def Rvir(self):
        Rvir_property = self.__profile_properties['Rvir_property']
        Rvir_f  = self.__halo_properties[Rvir_property]
        Rvir_unit = pg.UnitScalar(Rvir_f, self.__snapshot['pos'].units, subs=self.__snapshot)
        return Rvir_unit

    @property
    def gx_radius(self):
        gx_rad = self.__profile_properties['gx_radius']
        return gx_rad

    @property
    def Rgx(self):
        Rgx_unit = self.Rvir * self.gx_radius
        return Rgx_unit

    @property
    def center(self):
        return self.__center

    # #################################
    # load one File into np.arrays
    # #################################

    def __get_halo_filename(self):
        postfix_halo = self.__filename + '-halo.info'
        return self.__destination + "/" + postfix_halo

    def __get_gx_filename(self):
        postfix_gx = self.__filename + '-gx.info'
        return self.__destination + "/" + postfix_gx

    def __get_profile_filename(self):
        postfix_profile = 'profile-' + self.__profile + '.info'
        if self.__global_profile:
            profile_path = pg.gadget.general['Profile_dir'] + '/' + postfix_profile
            profile_path = os.path.expandvars(profile_path)
            return profile_path
        else:
            return postfix_profile

    def get_profile_path(self):
        profile_path = self.__destination + '/'
        return profile_path

    def exists_in_cache(self):
        try:
            self.__halo_properties._read_info_file(self.__get_halo_filename())
            self.__gx_properties._read_info_file(self.__get_gx_filename())
            return True
        except Exception as e:
            return False


    def load_snapshot(self, useCache=True, forceCache=False,
                      loadBinaryProperties=False, loadSF=True, loadGasHistory=False, H_neutral_only=None):
        # type: (bool, bool) -> pg.snapshot
        filename = self.__data_file
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("*********************************************")
            print(("loading snapshot " + filename + "..."))
            print("*********************************************")
        #snapshot = pg.Snap(filename, load_double_prec=True)
        snapshot = pg.Snapshot(filename, load_double_prec=True, H_neutral_only=H_neutral_only)
        self.__snapshot = snapshot

        if 'findgxfast' in self.__profile_properties:
            findgxfast = self.__profile_properties['findgxfast']
        else:
            findgxfast = False

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("boxsize ", snapshot.boxsize)
            print("particles ", snapshot.parts)
            print("headers")
            for h in snapshot.headers()[0]:
                print("......", h, " = ", snapshot.headers()[0][h])
            print("units", snapshot.gadget_units)

        try:
            if not useCache or forceCache:
                raise FileNotFoundError
            # halo properties are found, re-tranlsate and re-orientate the halo, nothing else has to be done
            self.__halo_properties._read_info_file(self.__get_halo_filename())
            self.__gx_properties._read_info_file(self.__get_gx_filename())
            if 'center_in_halo' in self.__gx_properties:
                center_known = self.__gx_properties['center_in_halo']
            else:
                # only dummy profile exists, do not reprocess
                self.__loaded = True
                return self.__snapshot

            self._center_gx(center_known)
            self._gx_orientate()
            if loadBinaryProperties:        # only when reloading, as no binary data in generated automatically
                self.halo_properties.load_binary_properties()
                self.gx_properties.load_binary_properties()
                self.profile_properties.load_binary_properties()
        except Exception as e:
            halo = None
            gx = None
            if not findgxfast:
                halo, gx, halos = self._prepare_zoom(mode='auto', shrink_on='stars', ret_FoF=True)
                if halos is not None:
                    self._create_halo_properties(halos[0])
                else:
                    self._create_halo_properties(None, headeronly=True)
                if gx is not None:
                    self.__galaxy_all = gx
                    self.__galaxy = gx
            else:
                self._create_halo_properties(None)
                self._center_gx(None)
                self._gx_orientate()

            self._create_gx_properties()

            if useCache or forceCache:
                self.__halo_properties._write_info_file(self.__get_halo_filename(), force=forceCache)
                self.__gx_properties._write_info_file(self.__get_gx_filename(), force=forceCache)
                self.__profile_properties._write_info_file(self.__get_profile_filename(), force=forceCache)

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("*********************************************")
            print("loading complete Rvir=", self.Rvir)
            print("*********************************************")
        self.__loaded = True

        if (loadSF or loadGasHistory):
            if 'redshift' in self.__gx_properties:
                rs = round(float(self.__gx_properties['redshift']), 2)
                a = round(float(self.__gx_properties['a']), 2)
                if True or rs == 0.0:  # load only for z=0
                    if loadSF and "starformingfile" in self.__gx_properties:
                        sfname = self.__gx_properties['starformingfile']
                        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                            print("*********************************************")
                            print("loading star forming information from ", sfname)
                            print("*********************************************")
                        star_form_filename = self.get_profile_path() + sfname
                        self._fill_star_from_info(star_form_filename, a)

                        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                            print("*********************************************")
                            print("reloading derived quanitities")
                            print("*********************************************")
                        self.__snapshot.fill_derived_rules(clear_old=True)

                    if loadGasHistory: # and "gashistoryfile" in self.__gx_properties:
                        sfname = self.__gx_properties['gashistoryfile']
                        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                            print("*********************************************")
                            print("loading gas history information from ", sfname)
                            print("*********************************************")
                        gas_trace_filename = self.get_profile_path() + sfname
                        self._fill_gas_from_traced(self.__snapshot, gas_trace_filename)
        return self.__snapshot


    def read_cache(self, forceCache = False):
        self.__halo_properties._read_info_file(self.__get_halo_filename())
        self.__gx_properties._read_info_file(self.__get_gx_filename())
        return

    def write_cache(self, forceCache = False):
        self.__halo_properties._write_info_file(self.__get_halo_filename(), force=forceCache)
        self.__gx_properties._write_info_file(self.__get_gx_filename(), force=forceCache)
        self.__profile_properties._write_info_file(self.__get_profile_filename(), force=forceCache)
        return

    def clone_profile(self, newprofile, BaseOnly=True, includeGalaxyPlot = True, force=True):
        halo_filename = os.path.split(self.__get_halo_filename())
        halo_dir = os.path.split(halo_filename[0])[0]
        halo_new = halo_dir + '/' + newprofile + '/' + halo_filename[1]

        if os.path.exists(halo_new) and not force:
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print("*****************************************************************")
                print("profile ", newprofile, "already exists - clone_profile terminated")
                print("*****************************************************************")
            return

        gx_filename = os.path.split(self.__get_gx_filename())
        gx_dir = os.path.split(gx_filename[0])[0]
        gx_new = gx_dir + '/' + newprofile + '/' + gx_filename[1]
        prop_filename = os.path.split(self.__get_profile_filename())
        prop_new = prop_filename[0] + '/profile-' + newprofile + '.info'

        if force:
            if os.path.exists(halo_new): os.remove(halo_new)
            if os.path.exists(gx_new): os.remove(gx_new)
            if os.path.exists(prop_new): os.remove(prop_new)

        #print("halo:", halo_new)
        #print("gx:", gx_new)
        #print("prop:", prop_new)
        if not os.path.exists(os.path.split(halo_new)[0]):
            os.makedirs(os.path.split(halo_new)[0])
        forceCache = True
        halo_props = ['date-created', 'profile-name', 'dm_only', 'mass', 'Mstars', 'Mgas', 'Mdm',
                      'parts', 'com', 'ssc', 'vel', 'vel_sigma', 'Rmax', 'lowres_part', 'lowres_mass',
                      'Rvir_FoF', 'R200_FoF', 'R200_ssc', 'M200_ssc', 'R200_com', 'M200_com',
                      'R500_FoF', 'R500_ssc', 'M500_ssc', 'R500_com', 'M500_com']

        gx_props = ['date-created', 'profile-name', 'Rvir_property', 'gx_radius', 'findgxfast',
                    'linking_length', 'linking_vel', 'lowres_threshold', 'a', 'cosmic_time', 'redshift',
                    'h', 'center_in_halo', 'center_com', 'center_ssc', 'center_halo_com', 'center_halo_ssc',
                    'R200', 'M200', 'R500', 'M500', 'M_stars', 'L_halo', 'L_baryons', 'M_gas',
                    'Z_stars', 'Z_gas', 'SFR', 'jzjc_baryons', 'jzjc_total', 'I_red',
                    'R_half_M', 'R_half_M_stellar', 'R_half_M(faceon)', 'R_eff(faceon)', 'D/T(stars)']

        #self.__profile_properties._write_info_file(prop_new, force=forceCache)
        new_prop_prop = SnapshotProperty()
        for p in self.__profile_properties:
            new_prop_prop.append(p, self.__profile_properties[p])
        new_prop_prop['profile-name'] = newprofile
        new_prop_prop._write_info_file(prop_new, force=forceCache)

        if BaseOnly:
            new_halo_prop = SnapshotProperty()
            for p in halo_props:
                new_halo_prop.append(p, self.__halo_properties[p])
            new_halo_prop['profile-name'] = newprofile
            new_gx_prop = SnapshotProperty()
            for p in gx_props:
                new_gx_prop.append(p, self.__gx_properties[p])
            new_gx_prop['profile-name'] = newprofile
            if includeGalaxyPlot:
                if 'galaxy-all' in self.__gx_properties:
                    new_gx_prop['galaxy-all'] = self.__gx_properties.get_binary_value('galaxy-all')
            new_halo_prop._write_info_file(halo_new, force=forceCache)
            new_gx_prop._write_info_file(gx_new, force=forceCache)
        else:
            self.__halo_properties['profile-name'] = newprofile
            self.__gx_properties['profile-name'] = newprofile
            self.__halo_properties.load_binary_properties()
            self.gx_properties.load_binary_properties()
            self.__halo_properties._write_info_file(halo_new, force=forceCache)
            self.__gx_properties._write_info_file(gx_new, force=forceCache)

        return


    def execute_command(self, command, par=None, par1=None, par2=None, par3=None, par4=None, par5=None):

        def load_command(script_file):
            command_str = ''
            if script_file == '':
                return command_str
            filename = script_file
            if len(script_file) > 3 and script_file[:-3].lower() != '.py':
                filename = filename + '.py'
            try:
                with open(filename, 'r') as myfile:
                    command_str = ''
                    ignore = True
                    intent = ''
                    for line in myfile:
                        if len(line) >= 3 and line[:3] == "def": ignore = False
                        if line.find('__name__') >= 0 and line.find('__main__'):
                            ignore = False
                        elif len(line) >= 15 and line[:15] == "# gcache3-init:":
                            ignore = True
                        elif len(line) >= 18 and line[:18] == "# gcache3-prepare:":
                            command_str += "if snap_exec == 'prepare':\n"
                            line = line.lstrip(' ')
                            intent = '    '
                            ignore = False
                        elif len(line) >= 18 and line[:18] == "# gcache3-process:":
                            command_str += "if snap_exec == 'process':\n"
                            line = line.lstrip(' ')
                            intent = '    '
                            ignore = False
                        elif len(line) >= 16 and line[:16] == "# gcache3-close:":
                            command_str += "if snap_exec == 'close':\n"
                            line = line.lstrip(' ')
                            intent = '    '
                            ignore = False
                        elif len(line) >= 14 and line[:14] == "# gcache3-end:":
                            ignore = True
                        elif len(line) >= 1 and line[:1] == "#":
                            line = ''
                        if not ignore and line != "":
                            command_str += intent + line
                    myfile.close()
            except Exception as e:
                print('*** error loading script ', filename)
                print(e)

            return command_str

        import numpy as np
        import matplotlib.pyplot as plt

        # processing as done in gCache, code should be shared instead of copied
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print('execute command on snapshot ' + self.__filename)
            print("*** execute command ", command)
            print("*** command parameter ", par)
            print('*** prepare')
        snap_num = 1
        snap_start = 1
        snap_end = 1
        cmd_par = par
        cmd_par1 = par1
        cmd_par2 = par2
        cmd_par3 = par3
        cmd_par4 = par4
        cmd_par5 = par5
        if not self.__loaded:
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print('*** load snapshot')
            self.load_snapshot()
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print('*** snapshot loaded')
        #dfSnap = self
        globalsParameter = globals().copy()
        globalsParameter['np'] = np
        globalsParameter['plt'] = plt

        localsParameter = locals().copy()
        localsParameter['snap_cache'] = self

        command_str = load_command(command)
        localsParameter['snap_exec'] = 'prepare'
        if command_str != '':
            exec(command_str, globalsParameter, localsParameter)

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print('*** process')
        localsParameter['snap_exec'] = 'process'
        if command_str != '':
            exec(command_str, globalsParameter, localsParameter)

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print('*** writing cache')
        self.write_cache()
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print('*** finished')

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print('*** close')
        localsParameter['snap_exec'] = 'close'
        if command_str != '':
            exec(command_str, globalsParameter, localsParameter)

        return

    def _load_profile(self, profile):
        self.__global_profile = False
        if profile == '':
            profile_name = 'cache'
        else:
            if len(profile) > 1:
                if profile[:1] == '$':
                    profile = profile[1:]
                    self.__global_profile = True
                else:
                    if not os.path.exists(profile):     # local profile not found
                        self.__global_profile = True    # switch to global mode
            profile_name = profile

        if self.__snap_home == '':
            split_names = os.path.split(self.__data_file)
            if split_names[0] != '':
                self.__destination = split_names[0] + '/' + profile_name
            else:
                self.__destination = profile_name
        else:
            split_names = os.path.split(self.__cache_home + self.__snap_name)
            if split_names[0] != '':
                self.__destination = split_names[0] + '/' + profile_name
            else:
                self.__destination = profile_name

        self.__filename = split_names[1]

        if not os.path.exists(self.__destination):
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print('create profle folder: "%s"' % self.__destination)
            os.makedirs(self.__destination)
        if os.listdir(self.__destination):
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print('WARNING: profile is not empty!')

        self.__profile = profile_name
        try:
            self.__profile_properties._read_info_file(self.__get_profile_filename())
        except Exception as e:
            # set default profile values
            self.__profile_properties = SnapshotProperty()
            self.__profile_properties.append("date-created", '"' + time.asctime(time.localtime(time.time())) + '"')
            self.__profile_properties.append("profile-name", profile_name)
            self.__profile_properties.append('Rvir_property', environment.DEFAULT_Rvir_property)
            self.__profile_properties.append('gx_radius', environment.DEFAULT_gx_radius)
            self.__profile_properties.append('findgxfast', environment.DEFAULT_findgxfast)
            self.__profile_properties.append('linking_length', environment.DEFAULT_linking_length)
            self.__profile_properties.append('linking_vel', environment.DEFAULT_linking_vel)
            self.__profile_properties.append('lowres_threshold', environment.DEFAULT_lowres_threshold)

        return

    def _correct_virial_info(self, halo_properties):
        def corr_factor(halo_properties, prop, factor):
            if not prop in halo_properties:
                return
            value = halo_properties[prop]
            value = value * factor
            halo_properties[prop] = value
            return

        mass_total = self.__snapshot['mass'].sum()
        mass_dm = self.__snapshot.dm['mass'].sum()
        factor = mass_total/mass_dm
        print("$$$Korrekturfaktor ", mass_total, mass_dm, factor)

        # Rvir_FoF: 299.3049442488739[ckpc h_0 ** -1]
        # R200_FoF: 287.7142082873453[ckpc  h_0 ** -1]
        # R200_ssc: 356.31469894527396[h_0 ** -1 ckpc]
        # M200_ssc: 1.051522e+03[1e+10 h_0 ** -1 Msol]
        # R200_com: 356.29890408445095[h_0 ** -1 ckpc]
        # M200_com: 1.051388e+03[1e+10 h_0 ** -1 Msol]
        # R500_FoF: 211.9896411873922[ckpc h_0 ** -1]
        # R500_ssc: 241.28093820131429[h_0 ** -1 ckpc]
        # M500_ssc: 816.2561863524752[1e+10 h_0 ** -1 Msol]
        # R500_com: 241.27632340717858[h_0 ** -1 ckpc]
        # M500_com: 816.2202291540016[1e+10 h_0 ** -1 Msol]

        corr_factor(halo_properties, 'Rvir_FoF', factor)
        corr_factor(halo_properties, 'R200_FoF', factor)
        corr_factor(halo_properties, 'R200_ssc', factor)
        corr_factor(halo_properties, 'M200_ssc', factor)
        corr_factor(halo_properties, 'R200_com', factor)
        corr_factor(halo_properties, 'M200_com', factor)
        corr_factor(halo_properties, 'R500_FoF', factor)
        corr_factor(halo_properties, 'R500_ssc', factor)
        corr_factor(halo_properties, 'M500_ssc', factor)
        corr_factor(halo_properties, 'R500_com', factor)
        corr_factor(halo_properties, 'M500_com', factor)
        return

    def _create_halo_properties(self, halo, headeronly=False):
        def convert_to_Units(prop):
            if isinstance(prop,tuple):
                value = pg.UnitArr([s for s in prop], units=None)
                return value
            else:
                return prop
        if not headeronly:
            if halo is not None:
                dm_halo = halo
                dm_only = False
            else:   # this part is used, when prepare_zoom is not used
                if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                    print("*********************************************")
                    print(("creating DM-halo ", self.__snapshot.parts))
                    print("*********************************************")

                search_snap = self.__snapshot.dm
                dm_only = True
                # old value 0.05
                if 'lowres_threshold' in self.__profile_properties:
                    threshold = self.__profile_properties['lowres_threshold']
                else:
                    threshold = pg.environment.DEFAULT_lowres_threshold
                # default old parameter linking_length=6 ckpc
                if 'linking_length' in self.__profile_properties:
                    linking_length = self.__profile_properties['linking_length']
                else:
                    linking_length = pg.environment.DEFAULT_linking_length

                if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                    print("search_snap ", search_snap.parts)

                if linking_length is None:
                    linking_length = (1500. * search_snap.cosmology.rho_crit(search_snap.redshift)
                         / np.median(search_snap['mass'])) ** pg.units.Fraction(-1, 3)

                    print("linking length assumed ", linking_length)

                FoF, N_FoF = pg.analysis.find_FoF_groups(search_snap, l=linking_length)
                halos = pg.analysis.generate_FoF_catalogue(search_snap, max_halos=1, FoF=FoF, progressbar=False,
                                                               exlude=lambda h, s: h.lowres_mass / h.mass > threshold)
                dm_halo = halos[0]  # most massive halo

        halo_properties = SnapshotProperty()
        halo_properties.append("date-created", '"' + time.asctime( time.localtime(time.time()) ) + '"')
        halo_properties.append("profile-name", self.__profile)

        if headeronly:
            self.__dm_halo = None
            self.__halo_properties = halo_properties
            return

        # create halo properties
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("DM-halo center ", dm_halo.ssc)
            print("DM-halo properties")

        halo_properties.append("dm_only", dm_only)
        for prop in dm_halo.props:
            v = convert_to_Units(dm_halo.props[prop])
            halo_properties[prop] = v
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print("......", prop, " = ", v)

        # if dm_only and dm_halo is not None:
        #     self._correct_virial_info(halo_properties)

        self.__dm_halo = dm_halo
        self.__halo_properties = halo_properties

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("*********************************************")
            print("creating complete ")
            print("*********************************************")
        return dm_halo

    def _create_gx_properties(self):

        info = SnapshotProperty()
        info.append("date-created", '"' + time.asctime( time.localtime(time.time()) ) + '"')
        info.append("profile-name", self.__profile)
        if self.__galaxy_all is None:
            self.__gx_properties = info
            return

        # to document calculations copy values to gx-properties
        # theses values should be used for recalculations instead of profile-values
        Rvir_property = self.__profile_properties['Rvir_property']
        gal_radius = self.__profile_properties['gx_radius']
        findgxfast = self.__profile_properties['findgxfast']
        linking_length = self.__profile_properties['linking_length']
        linking_vel = self.__profile_properties['linking_vel']
        lowres_threshold = self.__profile_properties['lowres_threshold']

        info.append('Rvir_property', Rvir_property)
        info.append('gx_radius', gal_radius)
        info.append('findgxfast',findgxfast)
        info.append('linking_length',linking_length)
        info.append('linking_vel',linking_vel)
        info.append('lowres_threshold',lowres_threshold)

        gx = self.__galaxy_all
        halo = self.__snapshot

        R200 = self.halo_properties['R200_com']
        M200 = self.halo_properties['M200_com']

        R500 = self.halo_properties['R500_com']
        M500 = self.halo_properties['M500_com']

        # infos part 1
        info['a'] = self.__snapshot.scale_factor
        info['cosmic_time'] = self.__snapshot.cosmic_time()
        info['redshift'] = self.__snapshot.redshift
        info['h'] = self.__snapshot.cosmology.h(self.__snapshot.redshift)
        #            print('center:             ', center, file=f)
        info['center_in_halo'] = self.__center           # the coord where the gx was found before translation
        info['center_com'] = pg.analysis.center_of_mass(gx) if gx is not None else None
        info['center_ssc'] = pg.analysis.shrinking_sphere(gx, center=[0] * 3,
                                                          R=R200) if gx is not None else None
        info['center_halo_com'] = pg.analysis.center_of_mass(halo) if halo is not None else None
        info['center_halo_ssc'] = pg.analysis.shrinking_sphere(halo, center=[0] * 3,
                                                               R=R200) if halo is not None else None
        info['R200'] = R200
        info['M200'] = M200
        info['R500'] = R500
        info['M500'] = M500

        # infos part 2
        # h = self.__snapshot[pg.BallMask(R200, center=self.__center, sph_overlap=False)]
        # h=halo?
        if halo:
            info['M_stars'] = halo.stars['mass'].sum().in_units_of('Msol', subs=self.__snapshot) \
                                        if len(halo.stars) else None
            info['L_halo'] = halo['angmom'].sum(axis=0)
            info['L_baryons'] = halo.baryons['angmom'].sum(axis=0)
        else:
            info['M_stars'] = pg.UnitArr(0, units='Msol')
            info['L_halo'] = None
            info['L_baryons'] = None
        if gx:
            info['M_stars'] = gx.stars['mass'].sum().in_units_of('Msol', subs=self.__snapshot)
            info['M_gas'] = gx.gas['mass'].sum().in_units_of('Msol', subs=self.__snapshot)
            info['Z_stars'] = (gx.stars['mass'] * gx.stars['metallicity']).sum() / gx.stars['mass'].sum() \
                                    if len(gx.stars) else None
            info['Z_gas'] =  (gx.gas['mass'] * gx.gas['metallicity']).sum() / gx.gas['mass'].sum()
            info['SFR'] = gx.gas['sfr'].sum()
            info['jzjc_baryons'] = gx.baryons['jzjc'].mean()
            info['jzjc_total'] = gx['jzjc'].mean()
            info['L_baryons'] = gx.baryons['angmom'].sum(axis=0)
            redI = pg.UnitArr(pg.analysis.reduced_inertia_tensor(gx)).flatten()
            info['I_red'] = redI
            info['R_half_M'] = pg.analysis.half_mass_radius(gx)
            info['R_half_M_stellar'] = pg.analysis.half_mass_radius(gx.stars)
        else:
            info['M_stars'] =  pg.UnitArr(0, 'Msol')
            info['M_gas'] = pg.UnitArr(0, 'Msol')
            info['Z_stars'] = None
            info['Z_gas'] = None
            info['SFR'] = None
            info['L_baryons'] = None
            info['I_red'] = None
            info['R_half_M'] = None

        # infos part 3
        disc_def = dict()  # jzjc_min=0.85, rmax='50 kpc', zmax='10 kpc')
        # pg.analysis.orientate_at(gx, 'red I', total=True)
        disc = gx[pg.DiscMask(**disc_def)]
        if gx:
            info['R_half_M(faceon)'] = pg.analysis.half_qty_radius(gx, 'mass', proj=2)
            info['R_eff(faceon)'] = pg.analysis.eff_radius(gx)
            info['D/T(stars)'] = float(disc.stars['mass'].sum() / gx.stars['mass'].sum())
        else:
            info['R_half_M(faceon)'] = None
            info['R_eff(faceon)'] = None
            info['D/T(stars)'] = None

        self.__gx_properties = info
        return


    def _center_gx(self, center_known):
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("*********************************************")
            print("center on galaxy...")
        stars = self.__snapshot.stars
        if center_known is None:
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print("searching for galaxies - shrinking sphere...")
            center = pg.analysis.shrinking_sphere(stars, center=[stars.boxsize / 2] * 3, R=stars.boxsize)
        else:
            center = center_known

        self.__center = center
        pg.Translation(-center).apply(self.__snapshot, total=True)

        vel_center = pg.analysis.mass_weighted_mean(self.__snapshot[self.__snapshot['r'] < '1 kpc'], 'vel')
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('center velocities at:', vel_center)
        self.__snapshot['vel'] -= vel_center

        # Rvir_property = self.__snap_config.Rvir_property
        # Rvir = self.dm_halo.props[Rvir_property]
        Rgx = self.Rvir * self.gx_radius
        #Rgx_units = pg.UnitScalar(Rgx).in_units_of(self.__snapshot['r'].units)
        #mask = pg.BallMask(str(Rgx) + ' ckpc h_0**-1', center=[0, 0 , 0])
        #mask = pg.BallMask(pg.UnitScalar(Rgx).in_units_of(self.snapshot['r'].units), center=[0, 0 , 0])
        mask = pg.BallMask(Rgx, center=[0, 0 , 0])
        snapshot_gx = self.__snapshot[mask]

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("center on galaxy - complete")
            print("*********************************************")

        self.__galaxy = snapshot_gx
        self.__galaxy_all = self.__galaxy
        return


    def _gx_orientate(self):
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("*********************************************")
            print("orientate galaxy...")

        #pg.analysis.orientate_at(self.__galaxy_all, 'L', total=True)
        if 'I_red' in self.__gx_properties:
            # redI = self.__gx_properties['I_red']
            redI = pg.UnitArr(pg.analysis.reduced_inertia_tensor(self.__galaxy_all)).flatten()
            if redI is not None:
                redI = redI.reshape((3, 3))
                mode, qty = 'red I', redI
                pg.analysis.orientate_at(self.__galaxy_all, mode, qty=qty, total=True)
            else:
                pg.analysis.orientate_at(self.__galaxy_all, 'L', total=True)
        else:
            pg.analysis.orientate_at(self.__galaxy_all, 'L', total=True)

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("orientate complete")
            print("*********************************************")
        return

    def _prepare_zoom(self, mode='auto', shrink_on='stars', ret_FoF=False,
                     sph_overlap_mask=False, star_form='deduce',
                     gas_trace='deduce', to_physical=True, load_double_prec=False,
                     fill_undefined_nan=True, gas_traced_blocks='all',
                     gas_traced_dervied_blocks=None, **kwargs):
        '''
        A convenience function to load a snapshot from a zoomed-in simulation that is
        not yet centered or oriented.

        Args:
            self (str, Snap):   The snapshot of a zoomed-in simulation to prepare.
                                Either as an already loaded snapshot or a path to the
                                snapshot.
            mode (str):         The mode in which to prepare the snapshot. You can
                                choose from:
                                    * 'auto':   try 'info', if it does not work
                                                fallback to 'FoF'
                                    * 'info':   read the info file (from `info`) and
                                                take the center and the reduced
                                                inertia tensor (for orientation) with
                                                fallback to angular momentum from this
                                                file
                                    * 'ssc':    center the snapshot using a
                                                shrinking sphere method (shrinking on
                                                the sub-snapshot specified by
                                                `shrink_on`, see below)
                                                ATTENTION: can easily center on a
                                                non-resolved galaxy!
                                    * 'FoF':    Do a FoF search on all particles and
                                                generate a halo catalogue for all
                                                those halos that have at most 1% mass
                                                from "lowres" particles. Then find the
                                                most massive FoF group of particles
                                                defined by `shrink_on` therein and
                                                finally perform a shrinking sphere on
                                                them to find the center. If
                                                `shrink_on` is 'all' or 'highres',
                                                skip the second FoF search and do the
                                                "ssc" on the most massive FoF group of
                                                the first step. The linking length for
                                                (both) the FoF finder is given by
                                                `linking_length`.
            info (str, dict):   Now gx_properties
                                Path to info file or the dictionary as returned from
                                `read_info_file`.
                                However, if set to 'deduce', it is tried to deduce the
                                path to the info file from the snapshot filename: it
                                is assumed to be in a subfolder 'trace' that is in the
                                same directory as the snapshot and named
                                `info_%03d.txt`, where the `%03d` is filled with the
                                snapshot number. The latter is taken as the last three
                                characters of the first dot / the end of the filename.
            shrink_on (str, list):
                                Define the part of the snapshot to perform the
                                shrinking sphere on, if there is no info file found or
                                info==None. It can be 'all' (use the entire
                                (sub-)snapshot `s`), a family name, or a list of
                                particle types (e.g. [0,1,4]).
            linking_length (UnitScalar):
                                The linking length used for the FoF finder in mode
                                "FoF" (if None, it defaults to
                                ( 1500.*rho_crit / median(mass) )^(-1/3) as defined in
                                `generate_FoF_catalogue`).
            linking_vel (UnitScalar):
                                The linking velocity used for the FoF finder in mode
                                "FoF". Only used for defining the galaxy, not for the
                                halo.
            ret_FoF (bool):     Also return the FoF group catalogue created (if in
                                mode "FoF").
            sph_overlap_mask (bool):
                                Whether to mask all particles that overlap into the
                                halo region (that is also include SPH particles that
                                are outside the virial radius, but their smoothing
                                length reaches into it).
            gal_R200 (float):   The radius to define the galaxy. Everything within
                                <gal_R200>*R200 will be defined as the galaxy.

            not clear how to use these parameters in first release, will be moved to cCache-commands
            star_form (str):    Path to the star formation file as written by the
                                program `gtrace`.
                                If set to 'deduce', its path is tried to build from
                                the snapshot filename by taking its directory and
                                adding '/trace/star_form.ascii'.
            gas_trace (str):    Path to the tracing file as written by the program
                                `gtracegas`.
                                If set to 'deduce', its path is tried to build from
                                the snapshot filename by taking its directory and
                                looking for files '/../*gastrace*'.
            to_physical (bool): Whether to convert the snapshot to physical units.
            load_double_prec (bool):
                                Force to load all blocks in double precision.
                                Equivalent with setting the snapshots attribute.
            fill_undefined_nan (bool):
                                Fill star formation info for IDs not listed in the
                                `star_form` file. For more info see
                                `fill_star_from_info`.
            gas_traced_blocks (iterable, str):
                                The blocks to add with `fill_gas_from_traced`.
            gas_traced_dervied_blocks (bool):
                                Passed to `fill_gas_from_traced` as `add_derived`.
                                Defaults to `gas_traced_blocks=='all'` if None.
            kwargs:             Passed to `generate_FoF_catalogue` in `mode='FoF'`.

        Returns:
            s (Snap):           The prepared snapshot.
            halo (SubSnap):     The cut halo of the found structure.
            gal (SubSnap):      The central galaxy of the halo as defined by
                                `gal_R200`.
        '''

        def get_shrink_on_sub(snap, shrink_on):
            if isinstance(shrink_on, str):
                shrink_on = str(shrink_on)
                if shrink_on == 'all':
                    return snap
                else:
                    return getattr(s, shrink_on)
            elif isinstance(shrink_on, list):
                return s[shrink_on]
            else:
                raise ValueError('`shrink_on` must be a family name or a list ' + \
                                 'of particle types, but was: %s' % (shrink_on,))

        def FoF_exclude(h, s, threshold=1e-2):
            if 'lowres_threshold' in self.__profile_properties:
                threshold = self.__profile_properties['lowres_threshold']
            else:
                threshold = pg.environment.DEFAULT_lowres_threshold

            M_lowres = h.lowres_mass
            M = h.mass
            return M_lowres / M > threshold

        # if 'gastrace' in kwargs:
        #     raise ValueError("You passed 'gastrace'. Did you mean 'gas_trace'?")
        # if 'starform' in kwargs:
        #     raise ValueError("You passed 'starform'. Did you mean 'star_form'?")

        s = self.__snapshot

        # if isinstance(s, str):
        #     s = Snapshot(s, load_double_prec=load_double_prec)
        gal_R200 = self.__profile_properties['gx_radius']
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print('prepare zoomed-in', s)

        # read info file (if required)
        if mode in ['auto', 'info']:
            if len(self.__gx_properties) == 0:
                if mode == 'auto':
                    mode = 'FoF'
                else:
                    raise IOError('Could not read/find the info file (gx.info)!')
            else:
                if mode == 'auto':
                    mode = 'info'

        # if to_physical:
        #     s.to_physical_units()

        # find center
        if mode == 'info':      # info file found an loaded into gx_properties
            center = self.__gx_properties['center']
        elif mode in ['ssc', 'FoF']:
            # default old parameter linking_length=None
            if 'linking_length' in self.__profile_properties:
                linking_length = self.__profile_properties['linking_length']
            else:
                linking_length = pg.environment.DEFAULT_linking_length

            # default old parameter linking_vel='200 km/s',
            if 'linking_vel' in self.__profile_properties:
                linking_vel = self.__profile_properties['linking_vel']
            else:
                linking_vel = pg.environment.DEFAULT_linking_vel

            if mode == 'FoF':
                try:
                    halos = pg.analysis.generate_FoF_catalogue(
                        s,
                        l=linking_length,
                        exclude=FoF_exclude,
                        #calc=['mass', 'lowres_mass'],       # generate all halo properties
                        max_halos=1,                         #  take only 3 instead of 10, for performance
                        progressbar=False,
                        **kwargs
                    )
                except Exception as e:
                    print("error finding halo ", e)
                    halos = None
                    halo = None
                    gal = None
                    if ret_FoF:
                        return halo, gal, halos
                    else:
                        return halo, gal

                if shrink_on not in ['all', 'highres']:     # $$ shrink_on parameter nicht klar
                    try:
                        galaxies = pg.analysis.generate_FoF_catalogue(
                            get_shrink_on_sub(s, shrink_on),
                            l=linking_length,
                            dvmax=linking_vel,
                            calc=['mass', 'com'],
                            max_halos=5,                    # take only 5 instead of 10
                            progressbar=False,
                            **kwargs
                        )
                        # The most massive galaxy does not have to be in a halo with little
                        # low resolution elements! Find the most massive galaxy living in a
                        # "resolved" halo:
                        galaxy = None
                        for gal in galaxies:
                            # since the same linking lengths were used for the halos and
                            # the galaxies (rethink that!), a galaxy in a halo is entirely
                            # in that halo or not at all
                            gal_ID_set = set(gal.IDs)
                            for h in halos:
                                if len(gal_ID_set - set(h.IDs)) == 0:
                                    galaxy = gal
                                    break
                            del h, gal_ID_set
                            if galaxy is not None:
                                break
                        del galaxies
                    except Exception as e:
                        galaxy = None
                        print("Error finding galaxy ", e)

                    if galaxy is None:
                        shrink_on = None
                    else:
                        shrink_on = s[galaxy]
                else:
                    shrink_on = s[halos[0]]
            elif mode == 'ssc':
                shrink_on = get_shrink_on_sub(s, shrink_on)

            if shrink_on is not None and len(shrink_on) > 0:
                com = pg.analysis.center_of_mass(s)
                R = np.max(pg.utils.periodic_distance_to(s['pos'], com, s.boxsize))
                center = pg.analysis.shrinking_sphere(shrink_on, com, R)
            else:
                center = None
        else:
            raise ValueError('Unkown mode "%s"!' % mode)

        # center in space
        if center is None:
            if environment.verbose >= environment.VERBOSE_TACITURN:
                print('no center found -- do not center')
        else:
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('center at:', center)
            pg.Translation(-center).apply(s, total=True)
            self.__center = center
            # center the velocities

        # cut the halo (<R200)
        if mode == 'info':
            R200 = self.__gx_properties['center']['R200']
            M200 = self.__gx_properties['center']['M200']
        else:
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('derive virial information')
            R200, M200 = pg.analysis.virial_info(s)
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('R200:', R200)
            print('M200:', M200)
        halo = s[pg.BallMask(R200, sph_overlap=sph_overlap_mask)]

        # orientate at the reduced inertia tensor of the baryons wihtin 10 kpc
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('orientate', end=' ')
        if mode == 'info':
            if 'I_red' in self.__gx_properties:
                redI = self.__gx_properties['I_red']
                if redI is not None:
                    redI = redI.reshape((3, 3))
                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print('at the galactic red. inertia tensor from info file')
                if environment.verbose >= environment.VERBOSE_TALKY:
                    print(redI)
                mode, qty = 'red I', redI
            else:
                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print('at angular momentum of the galaxtic baryons from info file:')
                mode, qty = 'vec', self.__gx_properties['L_baryons']
            s.orientate_at(s, mode, qty=qty, total=True)
        else:
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('at red. inertia tensor of the baryons within %.3f*R200' % gal_R200)
            pg.analysis.orientate_at(s[pg.BallMask(gal_R200 * R200, sph_overlap=False)].baryons,
                         'red I',
                         total=True
                         )

        # cut the inner part as the galaxy
        gal = s[pg.BallMask(gal_R200 * R200, sph_overlap=sph_overlap_mask)]
        Ms = gal.stars['mass'].sum()
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('M*:  ', Ms)

        if len(gal) == 0:
            gal = None
        if len(halo) == 0:
            halo = None

        # these functions will be moved to gCache-commands (former gtracegas, gtrace)
        # if star_form == 'deduce':
        #     try:
        #         star_form = os.path.dirname(s.filename) + '/trace/star_form.ascii'
        #     except:
        #         print('WARNING: could not deduce the path to the ' + \
        #               'star formation file!', file=sys.stderr)
        #         star_form = None
        # if isinstance(star_form, str):
        #     star_form = os.path.expanduser(star_form)
        #     if not os.path.exists(star_form):
        #         print('WARNING: There is no star formation file ' + \
        #               'named "%s"' % star_form, file=sys.stderr)
        #         star_form = None
        #     else:
        #         if environment.verbose >= environment.VERBOSE_NORMAL:
        #             print('read star formation file from:', star_form)
        #         fill_star_from_info(s, star_form,
        #                             fill_undefined_nan=fill_undefined_nan)
        #
        # if gas_trace == 'deduce':
        #     try:
        #         directory = os.path.dirname(s.filename) + '/../'
        #         candidates = []
        #         for fname in os.listdir(directory):
        #             if fname.startswith('gastrace'):
        #                 candidates.append(fname)
        #         if len(candidates) == 1:
        #             gas_trace = directory + candidates[0]
        #         else:
        #             raise RuntimeError('too many candidates!')
        #     except:
        #         print('WARNING: could not deduce the path to the ' + \
        #               'gas tracing file!', file=sys.stderr)
        #         gas_trace = None
        # if isinstance(gas_trace, str):
        #     gas_trace = os.path.expanduser(gas_trace)
        #     if not os.path.exists(gas_trace):
        #         print('WARNING: There is no gas trace file named ' + \
        #               '"%s"' % gas_trace, file=sys.stderr)
        #         gas_trace = None
        #     else:
        #         if environment.verbose >= environment.VERBOSE_NORMAL:
        #             print('read gas trace file from:', gas_trace)
        #         if gas_traced_dervied_blocks is None:
        #             gas_traced_dervied_blocks = (gas_traced_blocks == 'all')
        #         fill_gas_from_traced(s, gas_trace,
        #                              add_blocks=gas_traced_blocks,
        #                              add_derived=gas_traced_dervied_blocks)

        if mode == 'FoF' and ret_FoF:
            return halo, gal, halos
        else:
            return  halo, gal



    def _fill_star_from_info(self, fname, a_max, fill_undefined_nan=True, dtypes=None,
                            units=None):
        '''
        Read the formation radius rform and rform/R200(aform) from the star_form.ascii
        file and create the new blocks "rform" and "rR200form".

        Note:
            The set of the star particle IDs in the star formation information file
            must exactly be the set of star particle IDs in the (root) snapshot.

        Args:
            snap (Snap):    The snapshot to fill with the data (has to be the one at
                            z=0 of the simulation used to create the star_form.ascii).
            fname (str):    The path to the star_form.ascii file.
            fill_undefined_nan (bool):
                            Fill data with NaN for stars that are not listed in the
                            formation file.
            dtypes (np.dtypes):
                            The named(!) fields in the formation file. One of the has
                            to be a (integer type) field called "ID" for matching the
                            data to the snapshot.
            units (dict):   Specifying the units of the fields. Not every field has to
                            be sopecified here.
                            Defaults to: `{'aform':'a_form', 'rform':'kpc'}`.

        Raises:
            RuntimeError:   If the IDs are not unique or they do not match (except the
                            cases where `fill_undefined_nan` applies).
        '''
        snap = self.__snapshot
        stars = snap.root.stars
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print('reading the star formation information from %s...' % fname)
        # prepare the type of data
        if dtypes is None:
            dtypes = [('ID', np.uint64), ('aform', float), ('rform', float),
                      ('rR200form', float), ('Zform', float)]
        dtypes = np.dtype(dtypes)
        if 'ID' not in dtypes.fields:
            raise ValueError('The `dtypes` need to have a field "ID"!')
        if units is None:
            units = {'aform': 'a_form', 'rform': 'kpc'}
        try:
            # load the data
            SFI = np.loadtxt(fname, skiprows=1, dtype=dtypes)
            if a_max < 1.0:
                SFI = SFI[SFI['aform'] <= a_max]

            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print('testing if the IDs match the (root) snapshot...')
            SFI_IDs = SFI['ID']
            # test uniqueness removed
            # if not stars.IDs_unique():
            #    raise RuntimeError('Stellar IDs in the snapshot are not unique!')
            # if len(np.unique(SFI_IDs)) != len(SFI_IDs):
            #    raise RuntimeError('IDs in the star formation file are not unique!')
            # there might be too many or not enough IDs in the file

            missing = np.setdiff1d(stars['ID'], SFI_IDs, assume_unique=False)
            if fill_undefined_nan:
                if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                    print('WARNING: There are %d stellar IDs missing ' % len(missing) + \
                      'in the formation file! Fill them with NaN\'s.')
                add = np.array([(ID,) + (np.NaN,) * (len(dtypes) - 1) for ID in missing],
                               dtype=dtypes)
                SFI = np.concatenate((SFI, add))
                del add
                SFI_IDs = SFI['ID']
            else:
                raise RuntimeError('Some stars do not have a match in the ' + \
                                   'formation file "%s" (missing: %d)!' % (fname, len(missing))
                                   )
            toomany = np.setdiff1d(SFI_IDs, stars['ID'], assume_unique=False)
            if len(toomany):
                raise RuntimeError('Some formation file IDs do not have a match in ' + \
                                   'the snapshot (missing: %d)!' % (
                                       len(toomany))
                                   )

            if len(SFI) != len(stars):
                raise RuntimeError('Different number of stars in snapshot (%d) ' + \
                                   'and the formation file (%d)!' % (len(stars), len(SFI)))

            # adding the data as blocks
            sfiididx = np.argsort(SFI_IDs)
            sididx = np.argsort(stars['ID'])
            for name in dtypes.names:
                if name == 'ID':
                    continue
                if pg.environment.verbose >= pg.environment.VERBOSE_NORMAL:
                    print('adding the new block "%s" (units:%s, type:%s)...' % (
                        name, units.get(name, None), dtypes[name]))
                stars[name] = pg.UnitArr(np.empty(len(stars)),
                                         dtype=dtypes[name],
                                         units=units.get(name, None))
                stars[name][sididx] = SFI[name][sfiididx]
            return

        except Exception as e:
            print("error importing star_form file")
            print(e)
            exit(1)
            return

    def _fill_gas_from_traced(self, snap, data, add_blocks='all', add_derived=True,
                             units=None, invalid=0.0):
        '''
        Fill some information from the gas trace file into the snapshot as blocks.

        This function is using data from a gas trace file (as produced by `gtracegas`)
        to create new gas blocks with particle properties at in- and outflow times as
        well as additional information of the time being ejected.

        Among the new blocks are:
          "trace_type", "num_recycled",
          "infall_a", "infall_time",
          "mass_at_infall", "metals_at_infall", "jz_at_infall", "T_at_infall",
          "ejection_a", "ejection_time",
          "mass_at_ejection", "metals_at_ejection", "jz_at_ejection", "T_at_ejection",
          "cycle_r_max_at", "cycle_r_max", "cycle_z_max_at", "cycle_z_max",
        For the addtional derived blocks see `fill_derived_gas_trace_qty`.

        Note:
            The set of the star particle IDs in the star formation information file
            must exactly be the set of star particle IDs in the (root) snapshot.

        Args:
            snap (Snap):        The snapshot to fill with the data (has to be the one
                                at z=0 of the simulation used to create the trace
                                file).
            data (str, dict):   The path to the gas trace file or the alread read-in
                                data.
            add_blocks (iterable, str):
                                A list of the block names to actually add to the gas
                                sub-snapshot, or just 'all', which then adds all of
                                them. If `add_derived` it True, it will be set to 'all'
                                anyway, since they might be needed afterwards.
            add_derived (bool): Whether to also add blocks that can be derived from
                                the trace data stored in blocks. The function
                                `fill_derived_gas_trace_qty` is used for that.
            units (dict):       The units to use for masses, lengths, etc..
                                Defaults to: {'TIME':'a_form', 'MASS':'Msol',
                                              'TEMP':'K', 'ANGMOM':None, 'POS':'kpc'}
            invalid (float/NaN/inf):
                                The value to fill invalid entries with. Such entries
                                are properties of cycles that did not happen for a
                                given particle.
        '''

        def read_traced_gas(filename, types=None):
            '''
            Read the gas tracing statistics from a gtracegas output.

            The data also gets tagged by type:
                1:  gas in region
                2:  gas out of region
                3:  stars that formed in region
                4:  stars that formed outside region
            and the number of full cycles (leaving region + re-entering).

            TODO:
                * fill the stars that have been traced gas!

            The following is tracked at the different events:
                (re-)entering the region:       [a, mass, metals, j_z, T]
                the leave the region:           [a, mass, metals, j_z, T, vel]
                turning into a star:            [a, mass, metals, j_z]
                being out of the region, the following is updated:
                                                [(a, r_max), (a, z_max)]
            where a is the scale factor of when the particle reached the maximum radius or
            height, respectively.

            Args:
                filename (str):     The filename to read from.
                types (iterable):   The types of particles (see above!) to filter for.
                                    Default: all.

            Returns:
                tr (dict):          A dictionary with the trace data for the individual
                                    particles (of types ask for) indexed by ID. The actual
                                    data is:
                                        [ [type, n],
                                          np.ndarray of first enter event,
                                          np.ndarray leaving event,          \
                                          np.ndarray of traced being out,     >  n times
                                          np.ndarray of re-entering,         /
                                          events of leaving + out,          \  depending
                                          forming a star, or                 > on type or
                                          leaving + out + forming a stars,  / just nothing
                                        ]
                                    where n is the number of full recylces.
            '''
            filename = os.path.expanduser(filename)
            if pg.environment.verbose >= pg.environment.VERBOSE_TALKY:
                print('read gas trace file "%s"...' % filename)
                sys.stdout.flush()
            import pickle as pickle
            with open(filename, 'rb') as f:
                tr = pickle.load(f)
                f.close()
            return tr

        def fill_derived_gas_trace_qty(snap, units=None, invalid=0.0):
            """
            Derive blocks from the gas trace blocks and add them.

            Among the new blocks are:
              "out_time",
              "last_infall_a", "last_infall_time", "mass_at_last_infall",
                "metals_at_last_infall", "jz_at_last_infall", "T_at_last_infall",
              "last_ejection_a", "last_ejection_time", "mass_at_last_ejection",
                "metals_at_last_ejection", "jz_at_last_ejection", "T_at_last_ejection"'

            Note:
                The arguments `units` and `invalid` shall be the same as for
                `fill_gas_from_traced`.
                Also, some blocks added by `fill_gas_from_traced` are needed. Make sure
                they are avaiable!

            Args:
                snap (Snap):        The snapshot to fill with the data (has to be the one
                                    at z=0 of the simulation used to create the trace
                                    file).
                units (dict):       The units to use for masses, lengths, etc..
                invalid (float/NaN/inf):
                                    The value to fill invalid entries with. Such entries
                                    are properties of cycles that did not happen for a
                                    given particle.
            """
            if units is None:
                # TODO: angmom units
                units = dict(TIME='a_form', MASS='Msol', TEMP='K',
                             ANGMOM=None, POS='kpc')
            if pg.environment.verbose >= pg.environment.VERBOSE_NORMAL:
                print('adding blocks that can be derived from the gas trace blocks:')

            gas = snap.gas

            trmask = (gas['num_recycled'] != -1)
            set_N_cycles = set(gas['num_recycled'])
            max_N_cycle = max(set_N_cycles)

            # each (full) cycle takes some given time
            if max_N_cycle > 0:
                if pg.environment.verbose >= pg.environment.VERBOSE_NORMAL:
                    print('  "out_time",')
                ejected = gas['ejection_time'][:, :-1]
                infall = gas['infall_time'][:, 1:]
                gas['out_time'] = infall - ejected
                mask = (ejected == invalid) | ~np.isfinite(ejected)
                del ejected
                pg.environment.gc_full_collect()
                gas['out_time'][mask] = invalid
                mask = (infall == invalid) | ~np.isfinite(infall)
                del infall
                pg.environment.gc_full_collect()
                gas['out_time'][mask] = invalid
            pg.environment.gc_full_collect()

            # The events of the last infall and the last ejection are a bit messy to
            # access. Create extra blocks:
            # last_infall_idx = np.sum(~np.isnan(gas['infall_a']), axis=-1) - 1
            last_infall_idx = np.sum(gas['infall_a'] != invalid, axis=-1) - 1
            last_infall_idx = np.arange(len(last_infall_idx)), last_infall_idx
            for last, alle in [('last_infall_a', 'infall_a'),
                               ('last_infall_time', 'infall_time'),
                               ('mass_at_last_infall', 'mass_at_infall'),
                               ('metals_at_last_infall', 'metals_at_infall'),
                               ('jz_at_last_infall', 'jz_at_infall'),
                               ('T_at_last_infall', 'T_at_infall')]:
                if pg.environment.verbose >= pg.environment.VERBOSE_NORMAL:
                    print('  "%s"' % last)
                gas[last] = pg.UnitArr(np.empty(len(gas), dtype=gas[alle].dtype),
                                       units=gas[alle].units)
                gas[last][~trmask] = invalid
                gas[last][trmask] = gas[alle][last_infall_idx][trmask]
            del last_infall_idx
            pg.environment.gc_full_collect()

            # last_ejection_idx = np.sum(~np.isnan(gas['ejection_a']), axis=-1) - 1
            last_ejection_idx = np.sum(gas['ejection_a'] != invalid, axis=-1) - 1
            last_ejection_idx = np.arange(len(last_ejection_idx)), last_ejection_idx
            for last, alle in [('last_ejection_a', 'ejection_a'),
                               ('last_ejection_time', 'ejection_time'),
                               ('mass_at_last_ejection', 'mass_at_ejection'),
                               ('metals_at_last_ejection', 'metals_at_ejection'),
                               ('jz_at_last_ejection', 'jz_at_ejection'),
                               ('T_at_last_ejection', 'T_at_ejection')]:
                if pg.environment.verbose >= pg.environment.VERBOSE_NORMAL:
                    print('  "%s"' % last)
                gas[last] = pg.UnitArr(np.empty(len(gas), dtype=gas[alle].dtype),
                                       units=gas[alle].units)
                gas[last][~trmask] = invalid
                gas[last][trmask] = gas[alle][last_ejection_idx][trmask]
            del last_ejection_idx
            pg.environment.gc_full_collect()
            return

        def add_block(name, block):
            if name in add_blocks:
                if pg.environment.verbose >= pg.environment.VERBOSE_NORMAL:
                    print('  "%s"' % name)
                gas[name] = block
            return

        def count_cycles(history):
            c = 0
            for entry in history:
                if entry[0] == 'r':  # reentry
                    c += 1
            return c

        def get_a_at(event, history, max_count, invalid):
            a = []
            n = 0
            for entry in history:
                if entry[0] == event:
                    a.append(entry[1][0])  # a at event
                    n += 1
            while n < max_count:
                a.append(invalid)
                n += 1

            return a

        def get_event_Q(event, history, index_Q, max_count, invalid):
            Q = []
            n = 0
            for entry in history:
                if entry[0] == event:
                    Q.append(entry[1][index_Q])  # a at reentry
                    n += 1
            while n < max_count:
                Q.append(invalid)
                n += 1

            return Q

        if add_derived:
            add_blocks = 'all'
        if add_blocks == 'all':
            add_blocks = ["trace_type", "num_recycled",
                          "infall_a", "infall_time",
                          "mass_at_infall", "metals_at_infall", "jz_at_infall", "T_at_infall",
                          "ejection_a", "ejection_time",
                          "mass_at_ejection", "metals_at_ejection", "jz_at_ejection", "T_at_ejection",
                          "cycle_r_max_at", "cycle_r_max", "cycle_z_max_at", "cycle_z_max",
                          ]
        if isinstance(add_blocks, str):
            add_blocks = (add_blocks,)
        if units is None:
            # TODO: angmom units
            units = dict(TIME='a_form', MASS='Msol', TEMP='K',
                         ANGMOM=None, POS='kpc')
        gas = snap.root.gas
        pg.environment.gc_full_collect()

        if isinstance(data, str):
            filename = data
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print('reading the gas trace information from %s...' % filename)
            tr = read_traced_gas(data)
        else:
            filename = '<given data>'

        # filter to gas only (last history entry: new gas or reentered gas)
        # gas_type = {1, 2}
        # data = dict([i for i in iter(data.items()) if i[1][0][0] in gas_type])
        data = dict([i for i in tr.items() if i[1][len(i[1]) - 1][0] in ['n', 'r']])

        # print("root parts", snap.root.parts)
        # if len(set(data.keys()) - set(snap.root['ID'])) > 0: print("nicht in gesamt")
        # if len(set(data.keys()) - set(snap.root.stars['ID'])) > 0: print("nicht in stars")
        # if len(set(data.keys()) - set(snap.root.gas['ID'])) > 0: print("nicht in gas")

        allIDs = set(snap.root['ID'])
        todel = []
        for pk in data.keys():
            if not pk in allIDs:
                todel.append(pk)
        for pk in todel:
            data.pop(pk)

        if len(set(data.keys()) - set(snap.root['ID'])) > 0: print("nicht in gesamt")

        if pg.environment.verbose >= pg.environment.VERBOSE_TALKY:
            print('test IDs and find matching IDs...')
        if len(set(data.keys()) - set(gas['ID'])) > 0:
            raise RuntimeError('Traced gas IDs in "%s" have ' % filename +
                               '%s ' % pg.utils.nice_big_num_str(len(
                set(data.keys()) - set(gas['ID']))) +
                               'elements that are not in the snapshot!')
        tracedIDs = (set(data.keys()) & set(gas['ID']))

        trmask = np.array([(ID in tracedIDs) for ID in gas['ID']], dtype=bool)
        if pg.environment.verbose >= pg.environment.VERBOSE_TALKY:
            print('  found %s (of %s)' % (pg.utils.nice_big_num_str(len(tracedIDs)),
                                          pg.utils.nice_big_num_str(len(data))), end=' ')
            print('traced IDs that are in the snapshot')

        if pg.environment.verbose >= pg.environment.VERBOSE_NORMAL:
            print('adding the blocks:')

        trididx = np.argsort(list(data.keys()))
        gididx = np.argsort(gas['ID'])
        gididx_traced = gididx[trmask[gididx]]

        # type = not traced (0), in region (1), out of region (2)
        # number of full cycles (out and in)

        trace_type = pg.UnitArr(np.zeros(len(gas), dtype=int))
        trtype = np.array([1 if i[len(i) - 1][0] == 'n' else 2 for i in data.values()])
        trace_type[gididx_traced] = trtype[trididx]
        add_block('trace_type', trace_type)
        del trace_type

        n_cyc = np.array([count_cycles(i) for i in data.values()])
        num_recycled = pg.UnitArr(np.empty(len(gas), dtype=n_cyc.dtype))
        num_recycled[~trmask] = -1
        num_recycled[gididx_traced] = n_cyc[trididx]
        add_block('num_recycled', num_recycled)
        set_N_cycles = set(n_cyc)
        if pg.environment.verbose >= pg.environment.VERBOSE_TALKY:
            print('  +++ number of recycles that occured:', set_N_cycles, '+++')

        # Create blocks with shape (N, max(recycl)+1) for traced quantities at the
        # events of entering the region. N is the number of the gas particles. Each
        # particle has entries for each of their enter events, those events that do
        # not exists, get filled with `invalid`.

        max_N_cycle = max(set_N_cycles)
        # infall_t = np.array([[e[0] for e in i[1:1 + 3 * n + 1:3]] + [invalid] * (max_N_cycle - n)
        #                     for n, i in zip(n_cyc, list(data.values()))])
        infall_t = np.array([get_a_at('r', i, max_N_cycle + 1, invalid) for n, i in zip(n_cyc, list(data.values()))])
        infall_a = pg.UnitArr(np.empty((len(gas), max_N_cycle + 1),
                                       dtype=infall_t.dtype), units=units['TIME'])
        infall_a[~trmask] = invalid
        infall_a[gididx_traced] = infall_t[trididx]
        add_block('infall_a', infall_a)
        del infall_t
        pg.environment.gc_full_collect()

        # from .snapshot import age_from_form
        # only convert reasonable values & ensure not to overwrite blocks
        mask = (infall_a != invalid) & np.isfinite(infall_a)
        infall_time = infall_a.copy()
        new = gas.cosmic_time() - pg.age_from_form(infall_time[mask], subs=gas)
        infall_time.units = new.units
        infall_time[mask] = new
        del new
        add_block('infall_time', infall_time)
        del infall_time
        pg.environment.gc_full_collect()

        for name, idx, unit in [('mass_at_infall', 1, units['MASS']),
                                ('metals_at_infall', 2, units['MASS']),
                                ('jz_at_infall', 3, units['ANGMOM']),
                                ('T_at_infall', 4, units['TEMP'])]:
            if name not in add_blocks:
                continue
            # infall_Q = np.array([[e[idx] for e in i[1:1 + 3 * n + 1:3]] +
            #                     [invalid] * (max_N_cycle - n)
            #                     for n, i in zip(n_cyc, list(data.values()))])
            infall_Q = np.array(
                [get_event_Q('r', i, idx, max_N_cycle + 1, invalid) for n, i in zip(n_cyc, list(data.values()))])
            block = pg.UnitArr(np.empty((len(gas), max_N_cycle + 1),
                                        dtype=infall_Q.dtype),
                               units=unit)
            block[~trmask] = invalid
            block[gididx_traced] = infall_Q[trididx]
            add_block(name, block)
            del infall_Q, block
            pg.environment.gc_full_collect()

        # Create blocks with shape (N, max(recycl)+1) for traced quantities at the
        # events of ejection / leaving the region. N is the number of the gas
        # particles. Each particle has entries for each of their ejection events,
        # those events that do not exists, get filled with `invalid`.
        max_N_cycle = max(set_N_cycles)
        # eject_t = np.array([[e[0] for e in i[2:2 + 3 * n:3]] +
        #                     [i[2 + 3 * n][0] if t == 2 else invalid] +
        #                     [invalid] * (max_N_cycle - n)
        #                     for n, t, i in zip(n_cyc, trtype, list(data.values()))])
        eject_t = np.array([get_a_at('l', i, max_N_cycle + 1, invalid) for n, i in zip(n_cyc, list(data.values()))])

        ejection_a = pg.UnitArr(np.empty((len(gas), max_N_cycle + 1),
                                         dtype=eject_t.dtype),
                                units=units['TIME'])
        ejection_a[~trmask] = invalid
        ejection_a[gididx_traced] = eject_t[trididx]
        del eject_t
        add_block('ejection_a', ejection_a)
        pg.environment.gc_full_collect()

        # from .snapshot import age_from_form
        # only convert reasonable values & ensure not to overwrite blocks
        mask = (ejection_a != invalid) & np.isfinite(ejection_a)
        ejection_time = ejection_a.copy()
        new = gas.cosmic_time() - pg.age_from_form(ejection_time[mask], subs=gas)
        ejection_time.units = new.units
        ejection_time[mask] = new
        del new
        add_block('ejection_time', ejection_time)
        pg.environment.gc_full_collect()

        for name, idx, unit in [('mass_at_ejection', 1, units['MASS']),
                                ('metals_at_ejection', 2, units['MASS']),
                                ('jz_at_ejection', 3, units['ANGMOM']),
                                ('T_at_ejection', 4, units['TEMP'])]:
            if name not in add_blocks:
                continue
            # eject_Q = np.array([[e[idx] for e in i[2:2 + 3 * n:3]] +
            #                     [i[2 + 3 * n][idx] if t == 2 else invalid] +
            #                     [invalid] * (max_N_cycle - n)
            #                     for n, t, i in zip(n_cyc, trtype, list(data.values()))])
            eject_Q = np.array(
                [get_event_Q('l', i, idx, max_N_cycle + 1, invalid) for n, i in zip(n_cyc, list(data.values()))])
            block = pg.UnitArr(np.empty((len(gas), max_N_cycle + 1),
                                        dtype=eject_Q.dtype),
                               units=unit)
            block[~trmask] = invalid
            block[gididx_traced] = eject_Q[trididx]
            add_block(name, block)
            del eject_Q, block
            pg.environment.gc_full_collect()

        # for each cycle there is a maximum travel distance, plus one more for those
        # particles that are outside the region: store them
        for name, idx, unit in [('cycle_r_max_at', 0, units['TIME']),
                                ('cycle_r_max', 1, units['POS']),
                                ('cycle_z_max_at', 2, units['TIME']),
                                ('cycle_z_max', 3, units['POS'])]:
            if name not in add_blocks:
                continue
            # pos = np.array([[e[idx] for e in i[3:3 + 3 * n:3]] +
            #                 [i[3 + 3 * n][idx] if t == 2 else invalid] +
            #                 [invalid] * (max_N_cycle - n)
            #                 for n, t, i in zip(n_cyc, trtype, list(data.values()))])
            pos = np.array(
                [get_event_Q('p', i, idx, max_N_cycle + 1, invalid) for n, i in zip(n_cyc, list(data.values()))])
            block = pg.UnitArr(np.empty((len(gas), max_N_cycle + 1),
                                        dtype=pos.dtype),
                               units=unit)
            block[~trmask] = invalid
            block[gididx_traced] = pos[trididx]
            add_block(name, block)
            del pos, block
            pg.environment.gc_full_collect()

        if add_derived:
            fill_derived_gas_trace_qty(snap, units=units, invalid=invalid)
        return

    def print_results(self):
        print(" ")
        print("**********************************************************")
        print(("Summary of results for snapfile ", self.__data_file))
        print("**********************************************************")
        print(("number of results: ", len(self.results)))
        print("--------------------------------------------------------------")
        for result in self.results:
            print(("result:", result, "=", self.results[result]))
        print("--------------------------------------------------------------")
        return

class SnapshotProperty(dict):

    def __init__(self):
        dict.__init__(self)
        self.__modified = False
        self.__info_filename = ''

    def append(self, name, value):
        if name in self:
            v = self[name]
            if isinstance(v, pg.UnitArr) or isinstance(value, pg.UnitArr) or v != value:
                self[name] = value
                self.__modified = True
        else:
            v = None
            self[name] = value
            self.__modified = True
        return v        # returns old value

    def remove(self, name):
        if name in self:
            v = self[name]
            v = self.pop(name, None)
            self.__modified = True
        else:
            v = None
        return v

    @property
    def modified(self):
        return self.__modified


    def _write_info_file(self, info_filename, force=False):

        self.__info_filename = info_filename
        if not self.__modified and not force:
            return

        if os.path.exists(info_filename):
            os.remove(info_filename)

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("*********************************************")
            print('start writing SnapshotProperty file...')
            print("*********************************************")

        with open(info_filename, 'w') as f:
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print('file generated by gadutils.SnapshotFile', file=f)
            print('', file=f)
            infos = self
            for prop in infos:
                value = str(infos[prop]).replace('\n',' ')
                if isinstance(infos[prop], str):
                    value = str(infos[prop]).replace('\n', ' ')
                    value = '"' + value + '"'
                    print(prop, ' : ', value, file=f)
                elif isinstance(infos[prop], io.BytesIO):
                    buf = infos[prop]
                    buf.seek(0)
                    im = Image.open(buf)
                    save_parts = os.path.split(info_filename)
                    save_property = save_parts[1] + '.' + prop + '.png'
                    save_filename = save_parts[0] + '/' + save_property
                    im.save(save_filename)
                    buf.close()
                    print(prop, ' : ', '$' + save_property, file=f)
                else:
                    value = str(infos[prop]).replace('\n', ' ')
                    print(prop, ' : ', value, file=f)
            print('', file=f)
            f.close()
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("*********************************************")
            print('writing SnapshotProperty file complete')
            print("*********************************************")

        self.__modified = False
        return

    def get_binary_value(self, property):

        try:
            prop = self[property]
            if prop[0] != '$':
                return None
            prop = prop.strip('$')
            save_parts = os.path.split(self.__info_filename)
            save_filename = save_parts[0] + '/' + prop
            with open(save_filename, 'rb') as fin:
                data = io.BytesIO(fin.read())
                fin.close()
                return data
        except Exception as e:
            pass

        return None

    def load_binary_properties(self):
        loaded = 0
        for prop in self:
            value = self[prop]
            if isinstance(value,str) and value[0] == '$':
                bin_value = self.get_binary_value(prop)
                self[prop] = bin_value
                loaded += 1
        return loaded

    def _read_info_file(self, info_filename):
        '''
        Read in the contents of an info file as produced by gtrace into a dictionary.

        It is assumed, that there is exactly one colon per line, which separates the
        name (the part before the colon) and the value (the part after the colon). For
        the value part the following is assumed: if there are no brackets ('[') it is
        a float; if there is one pair of brackets, it is a float with units, which are
        given within the brackets; otherwise the value is simply stored as a string.

        Args:
            info_filename (str):     The path to the info file.

        Returns:
            info (dict):        A dictionary containing all the entries from the file.
        '''
        self.__info_filename = info_filename
        self.clear()
        info = self
        with open(os.path.expanduser(info_filename), 'r') as finfo:
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print("*********************************************")
                print("reading SnapshotProperty from file ", info_filename)
                print("*********************************************")
            for line in finfo:
                try:
                    cols = line.split(':', 1)
                    if len(cols) > 1:  # otherwise no value given, avoid ValueError
                        name = cols[0].strip()
                        value = cols[1].strip()
                        blocks = re.findall('\[(.*?)\]', value)
                        if len(blocks) == 1:
                            valuelist = len(value.rsplit('[')[0].strip()) == 0
                        else:
                            valuelist = False
                        if len(blocks) == 0:
                            if value[0] == '"':
                                value = value.strip('"')
                            elif value[0] == '$':
                                value = value.strip('"')
                            else:
                                if value == 'None':
                                    value = None
                                elif value == 'True':
                                    value = True
                                elif value == 'False':
                                    value = False
                                else:
                                    value = float(value)
                        elif len(blocks) == 1 and not valuelist:
                            value = pg.UnitArr(float(value.rsplit('[')[0].strip()),
                                            units=blocks[0])
                        elif len(blocks) == 1 and valuelist:
                            value = pg.UnitArr([int(s) if len(s.split('.')) == 1 else float(s) for s in value.strip('[]').split()],
                                               units=None)
                        elif len(blocks) == 2:
                            value = pg.UnitArr([float(s.strip(', ')) for s in blocks[0].split()],
                                            units=None if blocks[1] == 'None' else blocks[1])
                        if name in info:
                            print('WARNING: "%s" occures ' % name + \
                                  'multiple times in info file ' + \
                                  '"%s"! First one used.' % info_filename, file=sys.stderr)
                        info[name] = value
                except ValueError as e:
                    continue  # ignore error continue loading
            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print("SnapshotProperty")
            for prop in info:
                v = info[prop]
                if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                    print("......", prop, " = ", v)

            if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
                print("*********************************************")
                print("reading complete ")
                print("*********************************************")

        self.__modified = False
        return self



"""
Fit a spectrum with Voigt profiles, adding lines until the desired chisq value is achieved.  

Doctests:
    >>> import pygad as pg
    >>> ion = 'H1215'
    >>> s = pg.Snapshot(snap_file)
    >>> L = s.boxsize.in_units_of('Mpc', subs=s)
    >>> H = s.cosmology.H(s.redshift).in_units_of('(km/s)/Mpc', subs=s)
    >>> redshift = s.redshift
    >>> box_width = s.boxsize.in_units_of('Mpc', subs=s)
    >>> los = L.in_units_of('kpc') * np.random.rand(1,2)  # random LOS
    >>> lambda_rest = float(pg.analysis.absorption_spectra.lines[ion]['l'].split()[0])
    >>> spec_name = 'spec'+snap_file
    >>> vbox = H * box_width
    >>> v_limits = [0, vbox]
    >>> taus, col_densities, phys_densities, temps, metallicities, vpec, v_edges, restr_column = pg.analysis.absorption_spectra.mock_absorption_spectrum_of(s, los, line_name, v_limits, Nbins=(1+int(periodic_vel))*Nbins, return_los_phys=True)
    >>> velocities = 0.5 * (v_edges[1:] + v_edges[:-1])
    >>> wavelengths = lambda_rest * (s.redshift + 1) * (1 + velocities / c)
    >>> sigma_noise = 0.1  # add noise with this sigma (this gives S/N=10 per pixel)
    >>> noise = np.random.normal(0.0, sigma_noise, len(wavelengths))
    >>> noise_vector = np.asarray([sigma_noise] * len(noise))
    >>> fluxes = np.exp(-np.array(taus)) + noise
    >>> flux,noise_vector = pg.analysis.apply_LSF(wavelengths, fluxes, noise_vector, grating='COS_G130M')  # smooth with desired line spread fcn
    >>> contin = pg.analysis.fit_continuum(wavelengths, fluxes, noise_vector, order=0)  # do continuum fitting
    >>> fluxes = fluxes / contin
    >>> noise_vector  = noise_vector / contin
    >>> pg.analysis.write_spectrum(spec_name, los, lambda_rest, s.redshift, velocities, fluxes, taus, noise_vector, col_densities, phys_densities, temps, metallicities, vpec) # write spectrum
    >>> line_list = pg.analysis.fit_profiles(ion, lam, flux, noise, chisq_lim=2.0, max_lines=7, logN_bounds=[12,19], b_bounds=[3,100], mode='Voigt')
    >>> pg.analysis.write_lines(spec_name, line_list, 0) # append line info to spectrum file
    >>> model_flux, N, dN, b, db, l, dl, EW = pg.analysis.plot_fit(ax[ilos], lam, flux, noise, line_list, ion, starting_pixel=istart, show_plot=True)  # plot spectrum and lines
"""

__all__ = [
    "fit_profiles",
    "model_tau",
    "find_regions",
    "EquivalentWidth",
    "periodic_wrap",
    "periodic_unwrap_wavelength",
    "write_spectrum",
    "write_lines",
    "plot_fit",
]

from ..units import UnitArr
from ..physics import c

# from .. import utils
from .. import environment
import numpy as np
import pylab as plt

from .absorption_spectra import lines, Gaussian, Lorentzian, Voigt, line_profile


def fit_profiles(
    line,
    l,
    flux,
    noise,
    chisq_lim=2.0,
    max_lines=10,
    mode="Voigt",
    logN_bounds=[12, 19],
    b_bounds=[5, 200],
):
    """
    Fit Voigt/other profiles to the given spectrum.  Begins with one
    line, then adds lines until desired chi-sq is achieved.

    Args:
        line (str):         The line to fit as listed in
                            `analysis.absorption_spectra.lines`, e.g. 'HI1215'.
        l (array-like):     The wavelengths of the input spectrum to fit.
        flux (array-like):  The normalized flux at the given wavelengths,
                            i.e. the spectrum to fit.
        noise (array-like): The normalized 1-sigma noise vector at the given
                            wavelengths.  Must always be >0.
        chisq_lim (float):  Number of sigma below which chi-sq is considered
                            to be a "good fit" and no more lines are added.
                            If <0, then the value used is abs(chisq_lim)+0.1*n_lines,
                            where n_lines is the number of lines for that trial.
        max_lines (int):    Maximum number of lines allowed in a given detection
                            region, after which the fit declared done regardless of chisq.
                            If limit is hit, this may result in a poor fit.
        mode (str):         Type of line profile: Gaussian/Lorentzian/Voigt;
                            see absorption_spectra.line_profile().
        logN_bounds (list): Initial log(column density) is restricted to this range
        b_bounds (list):    Initial line width is restricted to this range (km/s)


    Returns:
        profiles:    Dictionary of [N, dN, b, db, l, dl, EW] of best-fit
                     profiles.
        tau_model:   Optical depths of best-fit model.

    """

    if isinstance(line, str):
        line = lines[line]
    l0 = line["l"]
    if isinstance(l, np.ndarray) or l.units in [
        1,
        None,
    ]:  # set units of l to Angstrom if none supplied
        l = UnitArr(l, "Angstrom")

    def _tau_to_flux(tau):  # return flux from tau, avoiding over/underflow
        return np.exp(-np.clip(tau, -50, 50))

    def _chisq(p, l, flux, noise, mode):  # reduced chisq
        return np.sum(
            (_tau_to_flux(model_tau(line, p, l, mode)) - flux) ** 2 / noise ** 2
        ) / len(l)

    def _add_line(p, bnd, l, flux, l0, mode):  # adds N, b, l for a new line
        l_bounds = [l[1], l[-2]]
        b_guess = (
            (l_bounds[1] - l_bounds[0]) / float(l0.split()[0]) * 3.0e5 / 5.0
        )  # first guess at b
        b_guess = max(2 * b_bounds[0], 0.5 * min(b_bounds[1], b_guess))
        if len(p) == 0:
            resid = flux
            n_guess = 14.0 - resid[np.argmin(resid)]  # first guess at logN
            p = np.array([n_guess])  # rough guess of logN
        else:
            resid = (
                1.0 + flux - _tau_to_flux(model_tau(line, p, l, mode))
            )  # residual spectrum
            n_guess = 14.0 - resid[np.argmin(resid)]  # first guess at logN
            p = np.append(p, n_guess)  # rough guess of logN
        p = np.append(p, b_guess)  # first guess of b
        p = np.append(p, l[np.argmin(resid)])  # add line @min of residual flux
        if len(bnd) == 0:
            bnd = np.array([logN_bounds])
        else:
            bnd = np.append(bnd, np.array([logN_bounds]), axis=0)
        bnd = np.append(bnd, np.array([b_bounds]), axis=0)
        bnd = np.append(bnd, np.array([l_bounds]), axis=0)
        # if environment.verbose <= environment.VERBOSE_TALKY:
        #    print('adding line logN=%g (%g) b=%g l=%g (%g-%g)'%(p[-3],min(resid),p[-2],p[-1],l[0],l[-1]))
        return p, bnd

    # identify independent regions to fit within the spectrum
    regions_l, regions_i = find_regions(l, flux, noise, min_region_width=2)

    # dicts to store results
    line_list = {
        "region": np.array([]),
        "l": np.array([]),
        "dl": np.array([]),
        "b": np.array([]),
        "db": np.array([]),
        "N": np.array([]),
        "dN": np.array([]),
        "EW": np.array([]),
    }

    # loop over regions
    from scipy.optimize import minimize

    for ireg in range(len(regions_l)):
        params = []
        bounds = []
        n_lines = 0
        best_nlines = 1
        chisq_old = 1.0e20
        chisq_accept = abs(chisq_lim)
        l_reg = l[regions_i[ireg, 0] : regions_i[ireg, 1]]
        f_reg = flux[regions_i[ireg, 0] : regions_i[ireg, 1]]
        n_reg = noise[regions_i[ireg, 0] : regions_i[ireg, 1]]
        # loop to add lines until desired chisq achieved
        while n_lines < max_lines:
            params, bounds = _add_line(params, bounds, l_reg, f_reg, l0, mode)
            n_lines = int(len(params) / 3)
            chisq_fcn = lambda *args: _chisq(*args)
            soln = minimize(
                chisq_fcn,
                params,
                bounds=bounds,
                args=(l_reg, f_reg, n_reg, mode),
                options={"maxiter": 100},
            )
            params = soln.x  # set params to new chisq-minimized values
            chisq_soln = _chisq(params, l_reg, f_reg, n_reg, mode)
            if chisq_soln < chisq_old:
                chisq_old = chisq_soln
                best_nlines = n_lines
            if environment.verbose >= environment.VERBOSE_TACITURN:
                print(
                    "Region %d: %d lines gives chisq=%g (%g) after %d iters"
                    % (ireg, n_lines, chisq_soln, chisq_accept, soln.nit)
                )
            """
            if params[n_lines*3-1] < l[2] or params[n_lines*3-1] > l[-2]:
                if environment.verbose <= environment.VERBOSE_TACITURN:
                    print('removing edge line %g<%g or %g>%g'%(params[n_lines*3-1], l[2], params[n_lines*3-1], l[-2]))
                params = params[:-3]  # if line is right at the edge, remove it and try again
                params += 0.02*(2*np.random.rand(len(params))-1)  # jiggle params to avoid hitting same solution
                chisq_soln = _chisq(params, l_reg, f_reg, n_reg, mode)
            """
            if chisq_soln < chisq_accept:
                break
            if chisq_lim < 0:
                chisq_accept += 0.1

        if chisq_soln > chisq_accept:  # try to go back to previous best solution
            if environment.verbose == environment.VERBOSE_TACITURN:
                print(
                    "Region %d uncoverged with %d lines, chisq=%g; retrying with %d"
                    % (ireg, n_lines, chisq_soln, best_nlines)
                )
            n_lines = best_nlines
            params = params[: 3 * n_lines]
            bounds = bounds[: 3 * n_lines]
            chisq_fcn = lambda *args: _chisq(*args)
            soln = minimize(
                chisq_fcn,
                params,
                bounds=bounds,
                args=(l_reg, f_reg, n_reg, mode),
                options={"maxiter": 100},
            )
            params = soln.x  # set params to new chisq-minimized values
            chisq_soln = _chisq(params, l_reg, f_reg, n_reg, mode)
            if chisq_soln > chisq_accept:
                if environment.verbose >= environment.VERBOSE_TACITURN:
                    print(
                        "WARNING: region %d has large chisq=%g; check fit"
                        % (ireg, chisq_soln)
                    )

        params += 0.02 * (
            2 * np.random.rand(len(params)) - 1
        )  # jiggle params and refit to compute hessian
        soln = minimize(
            chisq_fcn,
            params,
            args=(l_reg, f_reg, n_reg, mode),
            method="BFGS",
            options={"maxiter": 100},
        )
        cov = soln.hess_inv  # covariance matrix of final soluiton

        # append lines in this region onto line list
        if environment.verbose >= environment.VERBOSE_TACITURN:
            print(
                "region %d (%g-%g): chisq= %g with %d lines"
                % (
                    ireg,
                    regions_l[ireg, 0],
                    regions_l[ireg, 1],
                    chisq_soln,
                    int(len(params) / 3),
                )
            )
        for ip in range(n_lines):
            line_list["region"] = np.append(line_list["region"], ireg)
            line_list["N"] = np.append(line_list["N"], params[ip * 3])
            line_list["b"] = np.append(line_list["b"], params[ip * 3 + 1])
            line_list["l"] = np.append(line_list["l"], params[ip * 3 + 2])
            line_list["dN"] = np.append(line_list["dN"], np.sqrt(cov[ip * 3, ip * 3]))
            line_list["db"] = np.append(
                line_list["db"], np.sqrt(cov[ip * 3 + 1, ip * 3 + 1])
            )
            line_list["dl"] = np.append(
                line_list["dl"], np.sqrt(cov[ip * 3 + 2, ip * 3 + 2])
            )
            tau_line = model_tau(
                line,
                [params[ip * 3], params[ip * 3 + 1], params[ip * 3 + 2]],
                l_reg,
                mode,
            )
            line_list["EW"] = np.append(
                line_list["EW"], EquivalentWidth(_tau_to_flux(tau_line), l_reg)
            )

    return line_list


def model_tau(line, p, l, mode="Voigt"):
    """
    Compute optical depth vs. wavelength for a set of lines.

    Args:
        p (numpy array): [logN,b,wavelength] for a set of lines, in
                         a flattened numpy array.
        l (numpy array): Wavelengths over which to compute model spectrum.
    Returns:
        total_tau:  Optical depths (vs. l) from the combined set of lines
    """
    p = np.array(p)
    total_tau = np.zeros(len(l), dtype=float)
    if len(p) == 0:
        return total_tau  # no lines yet, return zeros
    for ip in range(int(len(p) / 3)):
        _, tau = line_profile(
            line, 10 ** p[ip * 3], b=p[ip * 3 + 1], l0=p[ip * 3 + 2], l=l, mode=mode
        )
        total_tau += tau  # add up optical depths from all lines
    return total_tau


def EquivalentWidth(fluxes, waves):
    """
    Find the equivalent width of a line/region.

    Args:
        taus (numpy array): the optical depths.
        waves (numpy array): list of wavelength for region.
    Returns:
        Equivalent width in units of waves.
    """
    dEW = np.zeros(len(fluxes))
    for i in range(1, len(fluxes) - 1):
        dEW[i] = (1.0 - fluxes[i]) * abs(waves[i + 1] - waves[i - 1]) * 0.5
    dEW[0] = (1.0 - fluxes[0]) * abs(waves[1] - waves[0])
    dEW[i - 1] = (1.0 - fluxes[i - 1]) * abs(waves[i - 1] - waves[i - 2])
    return np.sum(dEW)


def find_regions(
    wavelengths, fluxes, noise, min_region_width=2, N_sigma=5.0, extend=False
):
    """
    Finds detection regions above some detection threshold and minimum width.

    Args:
        wavelengths (numpy array)
        fluxes (numpy array): flux values at each wavelength
        noise (numpy array): noise value at each wavelength
        min_region_width (int): minimum width of a detection region (pixels)
        N_sigma (float): detection threshold (std deviations)
        extend (boolean): default is False. Option to extend detected regions untill tau
                        returns to continuum.

    Returns:
        regions_l (numpy array): contains subarrays with start and end wavelengths
        regions_i (numpy array): contains subarrays with start and end indices
    """

    num_pixels = len(wavelengths)
    pixels = range(num_pixels)
    min_pix = 1
    max_pix = num_pixels - 1

    flux_ews = [0.0] * num_pixels
    noise_ews = [0.0] * num_pixels
    det_ratio = [-float("inf")] * num_pixels

    # flux_ews has units of wavelength since flux is normalised. so we can use it for optical depth space
    for i in range(min_pix, max_pix):
        flux_dec = 1.0 - fluxes[i]
        if flux_dec < noise[i]:
            flux_dec = 0.0
        flux_ews[i] = 0.5 * abs(wavelengths[i - 1] - wavelengths[i + 1]) * flux_dec
        noise_ews[i] = 0.5 * abs(wavelengths[i - 1] - wavelengths[i + 1]) * noise[i]

    # dev: no need to set end values = 0. since loop does not set end values
    flux_ews[0] = 0.0
    noise_ews[0] = 0.0

    # Range of standard deviations for Gaussian convolution
    std_min = 2
    std_max = 11

    # Convolve varying-width Gaussians with equivalent width of flux and noise
    xarr = np.array([p - (num_pixels - 1) / 2.0 for p in range(num_pixels)])

    # this part can remain the same, since it uses EW in wavelength units, not flux
    for std in range(std_min, std_max):

        gaussian = np.exp(-0.5 * (xarr / std) ** 2)

        flux_func = np.convolve(flux_ews, gaussian, "same")
        noise_func = np.convolve(np.square(noise_ews), np.square(gaussian), "same")

        # Select highest detection ratio of the Gaussians
        for i in range(min_pix, max_pix):
            noise_func[i] = 1.0 / np.sqrt(noise_func[i])
            if flux_func[i] * noise_func[i] > det_ratio[i]:
                det_ratio[i] = flux_func[i] * noise_func[i]

    # Select regions based on detection ratio at each point, combining nearby regions
    start = 0
    region_endpoints = []
    for i in range(num_pixels):
        if start == 0 and det_ratio[i] > N_sigma and fluxes[i] < 1.0:
            start = i
        elif start != 0 and (det_ratio[i] < N_sigma or fluxes[i] > 1.0):
            if (i - start) > min_region_width:
                end = i
                region_endpoints.append([start, end])
            start = 0

    # made extend a kwarg option
    # lines may not go down to 0 again before next line starts

    if extend:
        # Expand edges of region until flux goes above 1
        regions_expanded = []
        for reg in region_endpoints:
            start = reg[0]
            i = start
            while i > 0 and fluxes[i] < 1.0:
                i -= 1
            start_new = i
            end = reg[1]
            j = end
            while j < (len(fluxes) - 1) and fluxes[j] < 1.0:
                j += 1
            end_new = j
            regions_expanded.append([start_new, end_new])

    else:
        regions_expanded = region_endpoints

    # Change to return the region indices
    # Combine overlapping regions, check for detection based on noise value
    # and extend each region again by a buffer
    regions_l = []
    regions_i = []
    buffer = 3
    for i in range(len(regions_expanded)):
        start = regions_expanded[i][0]
        end = regions_expanded[i][1]
        # TODO: this part seems to merge regions if they overlap - try printing this out to see if it can be modified to not merge regions?
        if i < (len(regions_expanded) - 1) and end > regions_expanded[i + 1][0]:
            end = regions_expanded[i + 1][1]
        for j in range(start, end):
            flux_dec = 1.0 - fluxes[j]
            if flux_dec > abs(noise[j]) * N_sigma:
                if start >= buffer:
                    start -= buffer
                if end < len(wavelengths) - buffer:
                    end += buffer
                regions_l.append([wavelengths[start], wavelengths[end]])
                regions_i.append([start, end])
                break

    if environment.verbose >= environment.VERBOSE_TACITURN:
        print("Found {} detection regions".format(len(regions_l)))
    return np.array(regions_l), np.array(regions_i)


def periodic_wrap(l, flux, noise):
    """
    To avoid situations where the end of a spectrum is in the middle of
    an absorption feature, this routine periodically wraps the spectrum
    such that the endpoint has the highest flux value.  The wavelengths
    are not changed, only the flux and noise vectors are wrapped.
    Should only be used with periodic simulations. Assumes the spectrum
    spans the entire simulation volume.

    Args:
        l (numpy array):     list of wavelength for region.
        flux (numpy array):  fluxes of wavelength for region.
        noise (numpy array): noise array (1sigma)

    Returns:
        flux:  periodically wrapped fluxes
        noise: periodically wrapped noise array
        starting_pixel: the pixel number where the highest flux occurs
    """

    starting_pixel = np.argmax(flux)
    flux = np.concatenate((flux[starting_pixel:-1], flux[0 : starting_pixel + 1]))
    noise = np.concatenate((noise[starting_pixel:-1], noise[0 : starting_pixel + 1]))
    if environment.verbose >= environment.VERBOSE_TACITURN:
        print("Periodically wrapping spectrum, starting_pixel= %d" % starting_pixel)

    return flux, noise, starting_pixel


def periodic_unwrap_wavelength(waves, l, starting_pixel):
    """
    After periodic_wrap(), the lines will have the wrong
    wavelength.  This routine 'unwraps' the wavelengths to
    place the line back where it should be.

    Args:
        waves (numpy array):  Wavelength values to be unwrapped.
        l (numpy array):      Wavelengths of the periodic spectrum.
        starting_pixel (int): The pixel number where the wrapping began.

    Returns:
        wave_line:  Values of wavelengths after unwrapping.
    """

    waves = waves - l[0] + l[starting_pixel]
    waves = np.where(waves > l[-1], waves - (l[-1] - l[0]), waves)  # wrap wavelengths

    return waves


def write_spectrum(
    spec_name,
    line,
    LOS_pos,
    lambda_rest,
    redshift,
    vels,
    fluxes,
    taus,
    noise,
    col_dens,
    phys_dens,
    temps,
    mets,
    vpec,
    overwrite=True,
):
    """
    Output spectrum to hdf5 file.

    Args:
        spec_name (str):      Name of file to write spectrum out to.
                              '.h5' will be appended to this.
        line (str):           The ion name, e.g. 'H1215'
        LOS_pos (list/array): (x,y,z) position of LOS,
                              with the LOS axis holding a value of -1.
        lambda_rest (float):  Rest wavelength of ion
        redshift (float):     Redshift of snapshot
        vels (list/array):    Velocities of pixels.
        fluxes (list/array):  Normalized fluxes of pixels; this should include
                              noise, smoothing, etc. so it's NOT =exp(-taus).
        taus (list/array):    Optical depths of pixels
        noise (list/array):   1-sigma noise array of pixels
        col_dens (list/array): Column densities for each pixel
        phys_dens (list/array): Tau-weighted physical densities for each pixel
        temps (list/array):   Tau-weighted gas temperatures for each pixel
        mets (list/array):    Tau-weighted metal mass fractions for each pixel
        vpec (list/array):    LOS peculiar velocity for each pixel

    Returns:

    """
    import h5py
    import os

    if os.path.isfile(spec_name) and not overwrite:
        if environment.verbose >= environment.VERBOSE_TACITURN:
            print(
                "WARNING: write_spectrum() failed: File %s exists, and overwrite set to False"
                % spec_name
            )
        return

    waves = lambda_rest * (redshift + 1.0) * (1.0 + vels / c)
    mets = np.log10(np.where(mets < 1.0e-10, 1.0e-10, mets))  # turn into log10(Z)
    if len(LOS_pos) == 2:
        LOS_pos = np.append(
            np.array(LOS_pos), -1.0
        )  # assumes if only 2 values are provided, they are (x,y), so we add -1 for z.

    with h5py.File("%s.h5" % spec_name, "w") as hf:
        lam0 = hf.create_dataset("lambda_rest", data=lambda_rest)
        lam0.attrs["ion_name"] = line  # store line name as attribute of rest wavelength
        hf.create_dataset("LOS_pos", data=np.array(LOS_pos))
        hf.create_dataset("redshift", data=redshift)
        hf.create_dataset("velocity", data=np.array(vels))
        hf.create_dataset("wavelength", data=np.array(waves))
        hf.create_dataset("flux", data=np.array(fluxes))
        hf.create_dataset("tau", data=np.array(taus))
        hf.create_dataset("noise", data=np.array(noise))
        hf.create_dataset("col_density", data=np.array(col_dens))
        hf.create_dataset("phys_density", data=np.array(phys_dens))
        hf.create_dataset("temperature", data=np.array(temps))
        hf.create_dataset("metallicity", data=np.array(mets))
        hf.create_dataset("vpec", data=np.array(vpec))

    return


def write_lines(spec_name, line_list, starting_pixel=0):
    """
    Append profile fit information to spectrum file.  Spectrum file must exist
    and contain the ion name attribute and wavelength list.

    Args:
        spec_name (str):      Name of file to write lines out to.
                              '.h5' will be appended to this.
                              Info placed in hdf5 group called 'lines'.
        line_list (dict):     List of fitted absorption feature
                              as output by fit_spectrum().
        starting_pixel (int): The pixel number where the wrapping began.
                              Set to 0 (default) for no periodic wrapping.

    Returns:

    """
    import h5py

    with h5py.File("%s.h5" % spec_name, "r") as hf:
        line = hf["lambda_rest"].attrs["ion_name"]
        waves = np.array(hf["wavelength"])

    # create overall model spectrum from all lines combined
    tau_model = np.zeros(len(waves))
    for i in range(len(line_list["N"])):
        p = np.array([line_list["N"][i], line_list["b"][i], line_list["l"][i]])
        tau_model += model_tau(line, p, waves)
    model_flux = np.exp(-np.clip(tau_model, -30, 30))

    # load data into arrays
    region = line_list["region"]
    N = line_list["N"]
    dN = line_list["dN"]
    b = line_list["b"]
    db = line_list["db"]
    if starting_pixel > 0:
        l = periodic_unwrap_wavelength(line_list["l"], waves, starting_pixel)
    else:
        l = dl = line_list["l"]
    dl = line_list["dl"]
    EW = line_list["EW"]

    with h5py.File("%s.h5" % spec_name, "a") as hf:
        lines = hf.create_group("lines")
        lines.create_dataset("fit_logN", data=np.array(N))
        lines.create_dataset("fit_dlogN", data=np.array(dN))
        lines.create_dataset("fit_b", data=np.array(b))
        lines.create_dataset("fit_db", data=np.array(db))
        lines.create_dataset("fit_l", data=np.array(l))
        lines.create_dataset("fit_dl", data=np.array(dl))
        lines.create_dataset("fit_EW", data=np.array(EW))
        lines.create_dataset("model_flux", data=np.array(model_flux))
        lines.create_dataset("starting_pixel", data=starting_pixel)

    return


def plot_fit(
    ax,
    waves,
    flux,
    noise,
    line_list,
    line,
    starting_pixel=0,
    show_plot=True,
    show_label=True,
):
    """
    Generates mdoel flux and plots spectrum, including the model flux and identified lines.  Returns model flux and line list.
    Args:
        waves (numpy array):  Wavelengths of original spectrum.
        flux (numpy array):   Fluxes of original spectrum.
        noise (numpy array):  Noise vector of original spectrum.
        line_list (dict):     List of fitted absorption feature
                              as output by fit_spectrum().
        line (str):           The ion name, e.g. 'H1215'
        starting_pixel (int): The pixel number where the wrapping began.
                              Set to 0 (default) for no periodic wrapping.

    Returns:
        flux_model:   Flux values of model (fitted) spectrum
        N, dN, b, db, l, dl, EW:  Fit parameters copied from line_list,
                      returned as numpy arrays.
    """
    tau_model = np.zeros(len(waves))
    for i in range(len(line_list["N"])):
        p = np.array([line_list["N"][i], line_list["b"][i], line_list["l"][i]])
        tau_model += model_tau(line, p, waves)
    flux_model = np.exp(-np.clip(tau_model, -30, 30))

    region = line_list["region"]
    N = line_list["N"]
    dN = line_list["dN"]
    b = line_list["b"]
    db = line_list["db"]
    if starting_pixel > 0:
        l = periodic_unwrap_wavelength(line_list["l"], waves, starting_pixel)
    else:
        l = dl = line_list["l"]
    dl = line_list["dl"]
    EW = line_list["EW"]

    if starting_pixel > 0:
        lamorig = periodic_unwrap_wavelength(waves, waves, starting_pixel)
    else:
        lamorig = waves
    sortid = np.argsort(lamorig)

    if show_label:
        ax.plot(lamorig[sortid], flux[sortid], alpha=0.4, label="flux")
        ax.plot(lamorig[sortid], noise[sortid], alpha=0.5, c="y")
        ax.plot(lamorig[sortid], flux_model[sortid], "--", label="model")
        ax.legend(loc="lower right")
    else:
        ax.plot(lamorig[sortid], flux[sortid], alpha=0.4)
        ax.plot(lamorig[sortid], noise[sortid], alpha=0.5, c="y")
        ax.plot(lamorig[sortid], flux_model[sortid], "--")
    for lw in l:
        ax.axvline(x=lw, ymin=0.95, ymax=1)

    if show_plot:
        plt.legend(loc="best")
        plt.show()

    return flux_model, N, dN, b, db, l, dl, EW



'''
Module for reading config files for Gadget parameters (not actual Gadget
configs!).

Example:
    >>> from ..environment import module_dir
    >>> read_config([module_dir+'config/gadget.cfg'])
    reading config file "gadget.cfg"
    >>> block_order
    ['POS ', 'VEL ', 'ID  ', 'MASS']
    >>> families.keys()
    dict_keys(['gas', 'stars', 'dm', 'bh', 'gands', 'sandbh', 'baryons', 'highres', 'lowres'])
    >>> families['dm']
    [1, 2, 3]
    >>> block_infos['MASS'], block_infos['HSML']
    ((1, 'float', None), (1, 'float', 'gas'))
    >>> get_block_units('RHO ').standardize()
    Unit("6.76991e-16 a**-3 g h_0**2 m**-3")
    >>> HDF5_to_std_name['Coordinates'], HDF5_to_std_name['ParticleIDs']
    ('POS ', 'ID  ')
    >>> std_name_to_HDF5['POS '], std_name_to_HDF5['ID  ']
    ('Coordinates', 'ParticleIDs')
    >>> elements[:3]
    ['He', 'C', 'Mg']
'''
__all__ = ['families', 'elements', 'default_gadget_units', 'block_units',
           'std_name_to_HDF5', 'HDF5_to_std_name', 'read_config',
           'get_block_units', 'general', 'block_infos']

from configparser import SafeConfigParser
from ..units import *
from os.path import exists, expanduser
from .. import environment
from .. import kernels
# already with some basic default values
families = {'gas':[0], 'stars':[4], 'dm':[1,2,3], 'bh':[5], 'baryons':[0,4,5]}
block_order = []
elements = []
# Horst: H_neutral_only added
general = {
    'kernel': '<undefined>',
    'vol_def_x': '<undefined>',
    'UVB': '<undefined>',
    'IMF': '<undefined>',
    'SSP_dir': '<undefined>',
    'unclear_blocks': 'warning',
    'Profile_dir': '',
    'H_neutral_only': False,
    }
block_infos = {}
# def. units have to be strings - they are used as replacements
default_gadget_units = {
    'LENGTH':   'ckpc/h_0',
    'VELOCITY': 'km/s',
    'MASS':     '1e10 Msol/h_0',
    }
block_units = {}
std_name_to_HDF5 = {}
HDF5_to_std_name = {}

def read_config(config):
    '''
    Reading some Gadget file definitions from a config file.

    The config file can have the following sections (the first two are required):

    families:           A definition of the families in terms of the particle
                        types. It must define gas, stars, dm (dark matter),
                        baryons, and bh (black holes).
    general:            A definition of the block ordering for format 1 files
                        (without info block), a list of the elements in block 'Z',
                        IFM definition, and more.
    base units:         The (default) Gadget base units (length, velocity, mass).
    block units:        The units for the different blocks.
    hdf5 names:         Name correspondences blocks in HDF5 files. (From HDF5 to
                        standard ones)

    Args:
        config (list):  list of possible filenames for the config file.
    '''
    global families, block_order, elements, default_gadget_units, block_units, \
           std_nameto_HDF5, HDF5_to_std_name, general, block_infos

    def test_section(cfg, section, entries):
        if not cfg.has_section(section):
            raise KeyError('Section "%s" is required in Gadget ' % section +
                           'config file.')
        if set(cfg.options(section)) < set(entries):
            raise ValueError('Section "%s" must have the ' % section +
                             'following entries: ' + str(entries))

    for filename in config:
        if exists(expanduser(filename)):
            break
    else:
        raise IOError('Config file "%s" does not exist!' % config)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        filenamep = os.path.split(filename)[1]
        print('reading config file "%s"' % filenamep)

    # The SafeConfigParser class has been renamed to ConfigParser in Python 3.2, not yet removed for compatibility
    cfg = SafeConfigParser(allow_no_value=True,
                           inline_comment_prefixes = ('#', ';') ) # new to python3. ignores comments at end of values
    cfg.optionxform = str
    cfg.read(filename)

    test_section(cfg, 'general', ['kernel', 'vol_def_x', 'IMF'])
    test_section(cfg, 'families', ['gas', 'stars', 'dm', 'bh', 'baryons'])
    test_section(cfg, 'base units', ['LENGTH', 'VELOCITY', 'MASS'])

    families.clear()
    for family in cfg.options('families'):
        if family in cfg.defaults():
            continue
        families[family] = sorted([int(t) for t
                in cfg.get('families',family).split(',')])

    while block_order: block_order.pop()
    if cfg.has_option('general', 'block order'):
        block_order += ['%-4s'%s.strip() for s in cfg.get('general','block order').split(',')]
    while elements: elements.pop()
    if cfg.has_option('general', 'elements'):
        elements += list(map(str.strip, cfg.get('general', 'elements').split(',')))
    kernel = cfg.get('general', 'kernel')
    if kernel not in kernels.kernels:
        raise ValueError('Kernel "%s" is unknown!' % kernel)
    general['kernel'] = kernel
    x = cfg.get('general', 'vol_def_x')
    if x == '1':
        x = 'ones(len(gas))'
    general['vol_def_x'] = x
    general['UVB'] = cfg.get('general', 'UVB')
    IMF = cfg.get('general', 'IMF')
    if IMF not in ['Kroupa', 'Salpeter', 'Chabrier']:
        raise ValueError('IMF "%s" is unknown!' % IMF)
    general['IMF'] = IMF
    if cfg.has_option('general', 'SSP_dir'):
        general['SSP_dir'] = cfg.get('general', 'SSP_dir',
                                     vars={'PYGAD_DIR':environment.module_dir})
    unclear_blocks = cfg.get('general', 'unclear_blocks')
    if unclear_blocks not in ['exception', 'warning', 'ignore']:
        raise ValueError('Unclear block mode "%s" is unknown!' % unclear_blocks)
    general['unclear_blocks'] = unclear_blocks

    if cfg.has_option('general', 'Profile_dir'):
        general['Profile_dir'] = cfg.get('general', 'Profile_dir')

    # Horst: read config-file into configuration
    if cfg.has_option('general', 'H_neutral_only'):
        general['H_neutral_only'] = cfg.getboolean('general','H_neutral_only')
    else:
        general['H_neutral_only'] = False # Gadget default value

    default_gadget_units.clear()
    default_gadget_units.update( cfg.items('base units') )

    block_infos.clear()
    if cfg.has_section('block infos'):
        from ast import literal_eval
        block_infos.update( { '%-4s'%n:literal_eval(u)
                                for n,u in cfg.items('block infos') } )

    block_units.clear()
    if cfg.has_section('block units'):
        block_units.update( { '%-4s'%n:u for n,u in cfg.items('block units') } )

    std_name_to_HDF5.clear()
    if cfg.has_section('hdf5 names'):
        std_name_to_HDF5.update( { '%-4s'%std:HDF5 for std,HDF5 \
                in cfg.items('hdf5 names') } )
    HDF5_to_std_name.clear()
    HDF5_to_std_name.update( { HDF5:std for std,HDF5 \
            in std_name_to_HDF5.items() } )

def get_block_units(block, gad_units=None):
    '''
    Return the default units (i.e. the units Gadget stores the block in) for a
    given block.

    Args:
        block (str):        The name of the block (e.g. "MASS").
        gad_units (dict):   The basic gadget units to use.

    Returns:
        unit (Unit):        The default units of this block.

    Raises:
        KeyError:           If the units are not known.
    '''
    if block not in block_units:
        raise KeyError('Units of block "%s" are not known' % block)
    if gad_units is None:
        gad_units = default_gadget_units
    gad_units = gad_units.copy()
    gad_units['TIME'] = gad_units['LENGTH'] + ' / (' + gad_units['VELOCITY'] + ')'

    u = block_units[block]
    for dimension, unit in gad_units.items():
        if isinstance(unit, str):
            u = u.replace(dimension, '('+str(unit)+')')
        else:
            u = u.replace(dimension, '('+str(Unit(unit))[1:-1]+')')
    return Unit(u).gather()



'''
Evaluate SPH properties.

Examples:
    >>> from ..environment import module_dir
    >>> from ..snapshot import Snapshot
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_320', physical=False)
    >>> from ..transformation import Translation
    >>> Translation([-34792.2, -35584.8, -33617.9]).apply(s)
    >>> s.to_physical_units()

    >>> if (abs(SPH_qty_at(s.gas, 'rho', [0,0,0]) - '1.74e5 Msol/kpc**3')
    ...             > '2e3 Msol/kpc**3'):
    ...     print(SPH_qty_at(s.gas, 'rho', [0,0,0]))
    load block pos... done.
    apply stored Translation to block pos... done.
    load block rho... done.
    load block hsml... done.
    load block mass... done.
    derive block dV... done.
    >>> if (abs(SPH_qty_at(s.gas, 'rho', s.gas['pos'][331798]) -
    ...         '5.33e4 Msol/kpc**3') > '2e2 Msol/kpc**3'):
    ...     print(SPH_qty_at(s.gas, 'rho', s.gas['pos'][331798]))
    >>> assert np.all(np.abs(SPH_qty_at(s, 'rho', s.gas['pos'][:1000:100]) -
    ...                      SPH_qty_at(s, 'rho', s.gas['pos'][:1000])[::100]) /
    ...                 SPH_qty_at(s, 'rho', s.gas['pos'][:1000:100]) < 1e-4)
    >>> scatter_gas_qty_to_stars(s, 'rho', name='gas_rho')  # doctest:+ELLIPSIS
    SimArr(...units="Msol kpc**-3", snap="snap_M1196_4x_320":stars)
    >>> if np.abs(s.stars['gas_rho'][0] - 3.5e7) > 0.2e7:
    ...     print(s.stars['gas_rho'][0])
    >>> if np.abs(s.stars['gas_rho'][-1] - 2.2e2) > 0.2e2:
    ...     print(s.stars['gas_rho'][-1])

    >>> v1 = SPH_qty_at(s, 'elements', s.gas['pos'][:200:10])
    load block Z... done.
    derive block elements... done.
    >>> v2 = SPH_qty_at(s, 'elements', s.gas['pos'][:200])[::10]
    >>> if np.max(np.abs(v1-v2)/v1) > 1e-6:
    ...     print(v1)
    ...     print(v2)

'''
__all__ = ['kernel_weighted', 'SPH_qty_at', 'scatter_gas_qty_to_stars']

import numpy as np
from ..units import *
from ..utils import dist


def _calc_ys(qty, all_gas_pos, gas_pos, hsml, kernel):
    y = np.empty((len(gas_pos),) + qty.shape[1:])
    for i in range(len(gas_pos)):
        hi = hsml[i]
        d = dist(all_gas_pos, gas_pos[i]) / hi
        mask = d < 1.0
        y[i] = np.sum((qty[mask].T * (kernel(d[mask]) / hi ** 3)).T,
                      axis=0)
    return y


def kernel_weighted(s, qty, units=None, kernel=None, parallel=None):
    '''
    Calculate a kernel weighted SPH quantity for all the gas particles:
             N
            ___         /  \
      y  =  \    x  W  | h  |
       i    /__   j  ij \ i/
            j=1

    Args:
        s (Snap):               The (sub-)snapshot to take the gas (quantity)
                                from.
        qty (str, array-like):  The name of the gas quantity or a array-like
                                object with length of the gas.
        units (str, Unit):      The units to convert the result in. If None, no
                                conversion is done.
        kernel (str):           The kernel to use. The default is to take the
                                kernel given in the `gadget.cfg`.

    Returns:
        y (UnitArr):            The kernel weigthed SPH property for all the gas
                                particles.
    '''
    # TODO: find ways to speed it up! -- parallelisation very inefficient
    gas = s.gas
    if isinstance(qty, str):
        qty = gas.get(qty)
    else:
        if len(qty) != len(gas):
            from ..utils import nice_big_num_str
            raise RuntimeError('The length of the quantity ' + \
                               '(%s) does not ' % nice_big_num_str(len(qty)) + \
                               'match the number of gas ' + \
                               '(%s)!' % nice_big_num_str(len(gas)))
    units = getattr(qty, 'units', None) if units is None else Unit(units)
    qty = np.asarray(qty)

    gas_pos = gas['pos'].view(np.ndarray)
    hsml = gas['hsml'].in_units_of(s['pos'].units, subs=s).view(np.ndarray)

    if kernel is None:
        from ..gadget import config
        kernel = config.general['kernel']
    from ..kernels import vector_kernels
    kernel = vector_kernels[kernel]

    if parallel is None:
        parallel = len(gas) > 1000
    if parallel:
        from multiprocessing import Pool, cpu_count
        N_threads = cpu_count()
        pool = Pool(N_threads)
        res = [None] * N_threads
        chunk = [[i * len(gas) / N_threads, (i + 1) * len(gas) / N_threads]
                 for i in range(N_threads)]
        import warnings
        with warnings.catch_warnings():
            # warnings.catch_warnings doesn't work in parallel
            # environment...
            warnings.simplefilter("ignore")  # for _z2Gyr_vec
            for i in range(N_threads):
                res[i] = pool.apply_async(_calc_ys,
                                          (qty, gas_pos,
                                           gas_pos[chunk[i][0]:chunk[i][1]],
                                           hsml[chunk[i][0]:chunk[i][1]],
                                           kernel))
        y = np.empty((len(gas_pos),) + qty.shape[1:])
        for i in range(N_threads):
            y[chunk[i][0]:chunk[i][1]] = res[i].get()
    else:
        y = _calc_ys(qty, gas_pos, gas_pos, hsml, kernel)

    return UnitArr(y, units)


def SPH_qty_at(s, qty, r, units=None, kernel=None, dV='dV'):
    '''
    Calculate a SPH quantity with the scatter approach at a given position.

    Args:
        s (Snap):               The (sub-)snapshot to take the gas (quantity)
                                from.
        qty (str, array-like):  The name of the gas quantity or a array-like
                                object with length of the gas.
        r (UnitQty):            The position(s) to evaluate the SPH quantity at.
        units (str, Unit):      The units to conver to. If None, no conversion is
                                done and the results is gioven in the units
                                encountered.
        kernel (str):           The kernel to use. The default is to take the
                                kernel given in the `gadget.cfg`.
        dV (str, array-like):   The volume measure of the SPH particles.

    Returns:
        Q (UnitArr):            The SPH property at the given position.
    '''
    r = UnitQty(r, s['pos'].units, subs=s)
    if not (r.shape == (3,) or (r.shape[1:] == (3,) and len(r.shape) == 2)):
        raise ValueError('Position `r` needs to have shape (3,) or (N,3)!')
    if isinstance(qty, str):
        qty = s.gas.get(qty)
    else:
        if len(qty) != len(s.gas):
            from ..utils import nice_big_num_str
            raise RuntimeError('The length of the quantity ' + \
                               '(%s) does not ' % nice_big_num_str(len(qty)) + \
                               'match the number of gas ' + \
                               '(%s)!' % nice_big_num_str(len(s.gas)))
    units = getattr(qty, 'units', None) if units is None else Unit(units)
    qty = np.asarray(qty)

    r = r.view(np.ndarray)
    gas_pos = s.gas['pos'].view(np.ndarray)
    hsml = s.gas['hsml'].in_units_of(s['pos'].units, subs=s).view(np.ndarray)
    dV = s.gas.get(dV).in_units_of(s['pos'].units ** 3, subs=s).view(np.ndarray)

    if kernel is None:
        from ..gadget import config
        kernel = config.general['kernel']
    from ..kernels import vector_kernels
    kernel_func = vector_kernels[kernel]

    if r.shape == (3,):
        d = dist(gas_pos, r) / hsml
        mask = d < 1.0
        Q = np.sum((qty[mask].T * (kernel_func(d[mask]) * dV[mask] /
                                   hsml[mask] ** 3)).T,
                   axis=0)
    elif len(r) < 100:
        dV_hsml3 = dV / hsml ** 3
        Q = np.empty((len(r),) + qty.shape[1:], dtype=qty.dtype)
        for i, x in enumerate(r):
            d = dist(gas_pos, x) / hsml
            mask = d < 1.0
            Q[i] = np.sum((qty[mask].T * (kernel_func(d[mask]) *
                                          dV_hsml3[mask])).T,
                          axis=0)
    else:
        from .. import C
        # C function expects doubles and cannot deal with views:
        r = r.astype(np.float64)
        gas_pos = gas_pos.astype(np.float64)
        hsml = hsml.astype(np.float64)
        dV = dV.astype(np.float64)
        qty = qty.astype(np.float64)
        if r.base is not None:
            r = r.copy()
        if gas_pos.base is not None:
            gas_pos = gas_pos.copy()
            hsml = hsml.copy()
            dV = dV.copy()
        # TODO: handle different types than double
        if len(qty.shape) == 1:
            Q = np.empty(len(r), dtype=np.float64)
            if qty.base is not None:
                qty = qty.copy()
            C.cpygad.eval_sph_at(
                C.c_size_t(len(r)),
                C.c_void_p(r.ctypes.data),
                C.c_void_p(Q.ctypes.data),
                C.c_size_t(len(gas_pos)),
                C.c_void_p(gas_pos.ctypes.data),
                C.c_void_p(hsml.ctypes.data),
                C.c_void_p(dV.ctypes.data),
                C.c_void_p(qty.ctypes.data),
                C.create_string_buffer(kernel.encode('ascii')),
                None  # build new tree
            )
        elif len(qty.shape) == 2:
            # C function needs contiquous arrays and cannot deal with
            # mutli-dimenensional ones...
            Q = np.empty((qty.shape[1], len(r)), dtype=np.float64)
            qty = qty.T
            if qty.base is not None:
                qty = qty.copy()
            # avoid the reconstruction of the octree
            from ..octree import cOctree
            tree = cOctree(gas_pos, hsml)
            for k in range(Q.shape[0]):
                C.cpygad.eval_sph_at(
                    C.c_size_t(len(r)),
                    C.c_void_p(r.ctypes.data),
                    C.c_void_p(Q[k].ctypes.data),
                    C.c_size_t(len(gas_pos)),
                    C.c_void_p(gas_pos.ctypes.data),
                    C.c_void_p(hsml.ctypes.data),
                    C.c_void_p(dV.ctypes.data),
                    C.c_void_p(qty[k].ctypes.data),
                    C.create_string_buffer(kernel.encode('ascii')),
                    C.c_void_p(tree._cOctree__node_ptr),
                )
            del tree
            Q = Q.T
        else:
            raise ValueError('Cannot handle more than two dimension in qty!')

    return UnitArr(Q, units)


def scatter_gas_qty_to_stars(s, qty, name=None, units=None, kernel=None, dV='dV'):
    '''
    Calculate a gas property at the positions of the stars and store it as a
    stellar property.

    This function calculates the gas property at the positions of the star
    particles by the so-called scatter approach, i.e. by evaluating the gas
    property of every gas particle at the stellar positions, kernel-weighted by
    the kernel of the gas particled the property comes from. Finally these
    quantities are stored as a new block for the stars of the snapshot. (Make
    shure there is no such block yet.)

    Args:
        s (Snap):               The snapshot to spread the properties of.
        qty (array-like, str):  The name of the block or the block itself to
                                spread onto the neighbouring SPH (i.e. gas)
                                particles.
        name (str):             The name of the new SPH block. If it is None and
                                `qty` is a string, is taken as the name.
        units (str, Unit):      The units to store the property in. If None, no
                                conversion is done and the results is stored in
                                the units encountered.
        kernel (str):           The kernel to use. The default is to take the
                                kernel given in the `gadget.cfg`.
        dV (str, array-like):   The volume measure of the SPH particles.

    Returns:
        Q (UnitArr):            The new SPH block.

    Raises:
        RuntimeError:       From this function directly:
                                If no name was given, though the qty is not a
                                string; or if the length of `qty` does not match
                                the number of stars in `s`.
                            From setting the new block:
                                If the length of the data does not fit the length
                                of the (sub-)snapshot; or if the latter is not the
                                correct "host" of the new block.
        KeyError:           If there already exists a block of that name.
    '''
    if name is None:
        if isinstance(qty, str):
            name = qty
        else:
            raise RuntimeError('No name for the quantity is given!')

    s.stars[name] = SPH_qty_at(s, qty=qty, r=s.stars['pos'], units=units, kernel=kernel)
    return s.stars[name]

